# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Sequence, TYPE_CHECKING

from numpy.typing import NDArray

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from pacman.model.placements import Placement

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)


class HasSynapses(object, metaclass=AbstractBase):
    """
    API for getting connections from the machine.
    """
    @abstractmethod
    def get_connections_from_machine(
            self, placement: Placement, app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> Sequence[NDArray]:
        """
        :param placement: Where the connection data is on the machine
        :param app_edge: The edge for which the data is being read
        :param synapse_info: The specific projection within the edge
        :returns: The connections from the machine for this vertex.
        """
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import struct
from spinn_utilities.log import FormatAdapter

logger = FormatAdapter(logging.getLogger(__name__))
_TWO_WORDS = struct.Struct("<II")


class EIEIOSpikeRecorder(object):
    """
    Records spikes using EIEIO format.
    """
    __slots__ = ("__record", )

    def __init__(self) -> None:
        self.__record = False

    @property
    def record(self) -> bool:
        """
        If the recorder is set to record
        """
        return self.__record

    @record.setter
    def record(self, new_state: bool) -> None:
        """
        Old method assumed to be spikes.
        """
        self.__record = bool(new_state)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol
from spynnaker.pyNN.external_devices_models.push_bot import (
    AbstractPushBotOutputDevice)


class PushBotMotor(AbstractPushBotOutputDevice):
    """
    The properties of the motor devices that may be set.
    The pushbot has two motors, 0 (left) and 1 (right).
    """
    # TODO: is that the right way round?

    #: For motor 0, set a particular speed
    MOTOR_0_PERMANENT = (
        0, MunichIoSpiNNakerLinkProtocol.push_bot_motor_0_permanent_key,
        -100, 100, 20
    )

    #: For motor 0, set a variable speed depending on time since event receive
    MOTOR_0_LEAKY = (
        1,
        (MunichIoSpiNNakerLinkProtocol
         .push_bot_motor_0_leaking_towards_zero_key),
        -100, 100, 20
    )

    #: For motor 0, set a particular speed
    MOTOR_1_PERMANENT = (
        2, MunichIoSpiNNakerLinkProtocol.push_bot_motor_1_permanent_key,
        -100, 100, 20
    )

    #: For motor 1, set a variable speed depending on time since event receive
    MOTOR_1_LEAKY = (
        3,
        (MunichIoSpiNNakerLinkProtocol
         .push_bot_motor_1_leaking_towards_zero_key),
        -100, 100, 20
    )



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Sequence

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged.range_dictionary import RangeDictionary

from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.models.neuron.threshold_types import AbstractThresholdType
from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.external_devices_models import (
    AbstractMulticastControllableDevice)

_DEVICE = "device"
_KEY = "key"
_SCALE = "scale"
_MIN = "min"
_MAX = "max"
_TS_INTER_SEND = "ts_inter_send"
_TS_NEXT_SEND = "ts_next_send"
_TYPE = "type"


class ThresholdTypeMulticastDeviceControl(AbstractThresholdType):
    """
    A threshold type that can send multicast keys with the value of
    membrane voltage as the payload.
    """
    __slots__ = ("__devices", )

    def __init__(self, devices: Sequence[AbstractMulticastControllableDevice]):
        """
        :param devices:
        """
        super().__init__(
            [Struct([
                (DataType.UINT32, _KEY),
                (DataType.UINT32, _SCALE),
                (DataType.S1615, _MIN),
                (DataType.S1615, _MAX),
                (DataType.UINT32, _TS_INTER_SEND),
                (DataType.UINT32, _TS_NEXT_SEND),
                (DataType.UINT32, _TYPE)])],
            {_KEY: "", _SCALE: "", _MIN: "mV", _MAX: "mV",
             _TS_INTER_SEND: "time steps", _TS_NEXT_SEND: "time steps",
             _TYPE: ""})
        self.__devices = devices

    @overrides(AbstractThresholdType.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[_KEY] = self._convert([
            d.device_control_key for d in self.__devices])
        parameters[_SCALE] = self._convert([
            d.device_control_scaling_factor for d in self.__devices])
        parameters[_MIN] = self._convert([
            d.device_control_min_value for d in self.__devices])
        parameters[_MAX] = self._convert([
            d.device_control_max_value for d in self.__devices])
        parameters[_TS_INTER_SEND] = self._convert([
            d.device_control_timesteps_between_sending
            for d in self.__devices])
        parameters[_TYPE] = self._convert([
            d.device_control_send_type.value for d in self.__devices])

    @overrides(AbstractThresholdType.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        state_variables[_TS_NEXT_SEND] = self._convert([
            d.device_control_first_send_timestep for d in self.__devices])



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Mapping, NamedTuple, Sequence, cast, TYPE_CHECKING
from spinn_utilities.overrides import overrides
from spinnman.model.enums import ExecutableType
from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM
from pacman.model.placements import Placement

from spinn_front_end_common.interface.provenance import (
    ProvidesProvenanceDataFromMachineImpl)
from spinn_front_end_common.interface.buffer_management.buffer_models import (
    AbstractReceiveBuffersToHost)
from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.interface.profiling.profile_utils import (
    get_profiling_data, reserve_profile_region, write_profile_region_data)
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary
from spinn_front_end_common.interface.buffer_management\
    .recording_utilities import (
        get_recording_header_size, get_recording_header_array)
from spinn_front_end_common.interface.simulation.simulation_utilities import (
    get_simulation_header_array)
from spinn_front_end_common.interface.ds import DataSpecificationGenerator
from spinn_front_end_common.interface.profiling import (
    AbstractHasProfileData, ProfileData)
from spinn_front_end_common.utilities.constants import SIMULATION_N_BYTES
if TYPE_CHECKING:
    from .population_vertex import PopulationVertex


class CommonRegions(NamedTuple):
    """
    Identifiers for common regions.
    """
    #: System control region
    system: int
    #: Provenance collection region
    provenance: int
    #: Profiling data region
    profile: int
    #: Recording channels region
    recording: int


# pylint: disable=abstract-method
class PopulationMachineCommon(
        MachineVertex,
        ProvidesProvenanceDataFromMachineImpl,
        AbstractReceiveBuffersToHost,
        AbstractHasProfileData,
        AbstractHasAssociatedBinary):
    """
    A common machine vertex for all population binaries.
    """

    __slots__ = (
        # Sdram used by the machine vertex
        "__sdram",
        # Regions to be used
        "__regions",
        # The total number of provenance items returned by this core
        "__n_provenance_items",
        # The profile tags to be decoded
        "__profile_tags",
        # The name of the binary to run on the core
        "__binary_file_name")

    def __init__(
            self, label: str, app_vertex: PopulationVertex,
            vertex_slice: Slice, sdram: AbstractSDRAM, regions: CommonRegions,
            n_provenance_items: int, profile_tags: Mapping[int, str],
            binary_file_name: str):
        """
        :param label: The label of the vertex
        :param app_vertex: The associated application vertex
        :param vertex_slice: The slice of the population that this implements
        :param sdram: The SDRAM used by the vertex
        :param regions: The regions to be assigned
        :param n_provenance_items:
            The number of additional provenance items to be read
        :param profile_tags:
            A mapping of profile identifiers to names
        :param binary_file_name: The name of the binary file
        """
        super().__init__(label, app_vertex, vertex_slice)
        self.__sdram = sdram
        self.__regions = regions
        self.__n_provenance_items = n_provenance_items
        self.__profile_tags = profile_tags
        self.__binary_file_name = binary_file_name

    @property
    @overrides(MachineVertex.sdram_required)
    def sdram_required(self) -> AbstractSDRAM:
        return self.__sdram

    @property
    def _pop_vertex(self) -> PopulationVertex:
        return cast('PopulationVertex', self._app_vertex)

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self) -> int:
        return self.__regions.provenance

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self) -> int:
        return self.__n_provenance_items

    @overrides(AbstractReceiveBuffersToHost.get_recording_region_base_address)
    def get_recording_region_base_address(self, placement: Placement) -> int:
        return locate_memory_region_for_placement(
            placement, self.__regions.recording)

    @overrides(AbstractHasProfileData.get_profile_data)
    def get_profile_data(self, placement: Placement) -> ProfileData:
        return get_profiling_data(
            self.__regions.profile, self.__profile_tags, placement)

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self) -> ExecutableType:
        return ExecutableType.USES_SIMULATION_INTERFACE

    def _write_common_data_spec(
            self, spec: DataSpecificationGenerator,
            rec_regions: Sequence[int]) -> None:
        """
        Write the data specification for the common regions.

        :param spec:
            The data specification to write to
        :param rec_regions:
            A list of sizes of each recording region (including empty ones)
        """
        # Write the setup region
        spec.reserve_memory_region(
            region=self.__regions.system, size=SIMULATION_N_BYTES,
            label='System')
        spec.switch_write_focus(self.__regions.system)
        spec.write_array(get_simulation_header_array(self.__binary_file_name))

        # Reserve memory for provenance
        self.reserve_provenance_data_region(spec)

        # Write profile data
        reserve_profile_region(
            spec, self.__regions.profile, self._pop_vertex.n_profile_samples)
        write_profile_region_data(
            spec, self.__regions.profile, self._pop_vertex.n_profile_samples)

        # Set up for recording
        spec.reserve_memory_region(
            region=self.__regions.recording,
            size=get_recording_header_size(len(rec_regions)),
            label="Recording")
        spec.switch_write_focus(self.__regions.recording)
        spec.write_array(get_recording_header_array(rec_regions))

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self) -> str:
        return self.__binary_file_name

    @overrides(MachineVertex.get_n_keys_for_partition)
    def get_n_keys_for_partition(self, partition_id: str) -> int:
        # Colour each time slot with up to 16 colours to allow for delays
        return self.vertex_slice.n_atoms * 16



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex
from pacman.model.placements import Placement

NEURON_EXPANDER_APLX = "neuron_expander.aplx"


@require_subclass(MachineVertex)
class AbstractNeuronExpandable(object, metaclass=AbstractBase):
    """
    Indicates a class (a
    :py:class:`~pacman.model.graphs.machine.MachineVertex`)
    that has may need to run the neuron expander APLX.
    """

    __slots__ = ()

    @abstractmethod
    def gen_neurons_on_machine(self) -> bool:
        """
        True if this vertex should be generated on the machine.

        .. note::
            The typical implementation for this method will be to ask the
            neuron data

        :returns: True if this vertex can generate its data on machine
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def neuron_generator_region(self) -> int:
        """
        The region containing the parameters of neuron expansion.
        """
        raise NotImplementedError

    @abstractmethod
    def read_generated_initial_values(self, placement: Placement) -> None:
        """
        Fill in any requested initial values.

        :param placement: Where the data is on the machine
        """
        raise NotImplementedError



# Copyright (c) 2023 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Types (and related) that are useful for implementing connectors.
"""

from typing import Final, Iterable, Optional, Sequence, List, Union

import numpy
from numpy.typing import NDArray
from typing_extensions import TypeAlias, TypeGuard

import neo
from pyNN.random import RandomDistribution

#: The type of weights and delays provided by Synapse / SynapseInformation
# Combined types (where value could be either)
WEIGHTS_DELAYS: Final['TypeAlias'] = Optional[Union[
    float, str, RandomDistribution, NDArray[numpy.float64]]]
WEIGHTS: Final['TypeAlias'] = Optional[Union[
    float, str, RandomDistribution, NDArray[numpy.float64]]]
DELAYS: Final['TypeAlias'] = \
    Union[float, str, RandomDistribution, NDArray[numpy.float64]]
# These are the Types we know are coming in.
# Most things that can be considered floats (including int)  will work
WEIGHTS_DELAYS_IN: Final['TypeAlias'] = Optional[Union[
    float, str, RandomDistribution, Iterable[float], NDArray[numpy.float64]]]

IoDest: TypeAlias = Union[
    str, neo.baseio.BaseIO, None]  # pylint: disable=no-member

ViewIndices = Union[None, Sequence[int], NDArray[numpy.integer]]
#: :meta private:
Selector: TypeAlias = Union[
    None, int, slice, Sequence[int], List[bool], NDArray[numpy.bool_],
    NDArray[numpy.integer]]

WeightScales: TypeAlias = Union[NDArray[numpy.floating], Sequence[float]]


def is_scalar(value: WEIGHTS_DELAYS) -> TypeGuard[Union[int, float]]:
    """
    Are the weights or delays a simple integer or float?

    :returns: True if the type of `value` is a scalar type.
    """
    return numpy.isscalar(value)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Tuple, Union

from numpy import integer, uint32
from numpy.typing import NDArray

from spinn_utilities.abstract_base import AbstractBase, abstractmethod

from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    ConnectionsArray)

from .abstract_sdram_synapse_dynamics import AbstractSDRAMSynapseDynamics


class AbstractPlasticSynapseDynamics(
        AbstractSDRAMSynapseDynamics, metaclass=AbstractBase):
    """
    Synapses which change over time.
    """

    __slots__ = ()

    @abstractmethod
    def get_n_words_for_plastic_connections(self, n_connections: int) -> int:
        """
        :param n_connections:
        :returns:
           The number of 32-bit words for `n_connections` in a single row.
        """
        raise NotImplementedError

    @abstractmethod
    def get_plastic_synaptic_data(
            self, connections: ConnectionsArray,
            connection_row_indices: NDArray[integer], n_rows: int,
            n_synapse_types: int,
            max_n_synapses: int, max_atoms_per_core: int) -> Union[
                Tuple[NDArray[uint32], NDArray[uint32],
                      NDArray[uint32], NDArray[uint32]],
                Tuple[List[NDArray[uint32]], List[NDArray[uint32]],
                      NDArray[uint32], NDArray[uint32]]]:
        """
        Get the fixed-plastic data, and plastic-plastic data for each row, and
        lengths for the fixed_plastic and plastic-plastic parts of each row.

        Data is returned as an array made up of an array of 32-bit words for
        each row, for each of the fixed-plastic and plastic-plastic data
        regions.  The row into which connection should go is given by
        `connection_row_indices`, and the total number of rows is given by
        `n_rows`.

        Lengths are returned as an array made up of an integer for each row,
        for each of the fixed-plastic and plastic-plastic regions.

        :param connections: The connections to get data for
        :param connection_row_indices:
            The row into which each connection should go
        :param n_rows: The total number of rows
        :param n_synapse_types: The number of synapse types
        :param max_n_synapses: The maximum number of synapses to generate
        :param max_atoms_per_core: The maximum number of atoms on a core
        :return: (fp_data (2D), pp_data (2D), fp_size (1D), pp_size (1D))
        """
        raise NotImplementedError

    @abstractmethod
    def get_n_plastic_plastic_words_per_row(
            self, pp_size: NDArray[uint32]) -> NDArray[integer]:
        """
        :returns: The number of plastic words to be read from each row.
        """
        raise NotImplementedError

    @abstractmethod
    def get_n_fixed_plastic_words_per_row(
            self, fp_size: NDArray[uint32]) -> NDArray[integer]:
        """
        :returns: The number of fixed plastic words to be read from each row.
        """
        raise NotImplementedError

    @abstractmethod
    def get_n_synapses_in_rows(
            self, pp_size: NDArray[uint32],
            fp_size: NDArray[uint32]) -> NDArray[integer]:
        """
        Get the number of synapses in each of the rows with plastic sizes
        `pp_size` and `fp_size`.

        :param pp_size:
        :param fp_size:
        :returns: The number of synapses in each of the rows
        """
        raise NotImplementedError

    @abstractmethod
    def read_plastic_synaptic_data(
            self, n_synapse_types: int,
            pp_size: NDArray[uint32], pp_data: List[NDArray[uint32]],
            fp_size: NDArray[uint32], fp_data: List[NDArray[uint32]],
            max_atoms_per_core: int) -> ConnectionsArray:
        """
        Read the connections indicated in the connection indices from the
        data in `pp_data` and `fp_data`.

        :param n_synapse_types:
        :param pp_size: 1D
        :param pp_data: 2D
        :param fp_size: 1D
        :param fp_data: 2D
        :param max_atoms_per_core:
        :return:
            array with columns ``source``, ``target``, ``weight``, ``delay``
        """
        raise NotImplementedError



# Copyright (c) 2024 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Sequence, Optional, TYPE_CHECKING

import numpy
from numpy import uint32
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.interface.ds import DataType
from spynnaker.pyNN.types import WEIGHTS

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_machine import (
    AbstractGenerateConnectorOnMachine, ConnectorIDs)
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import SynapseInformation
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge)


class AllButMeConnector(AbstractGenerateConnectorOnMachine,
                        AbstractGenerateConnectorOnHost):
    """
    A Connector that connect all the neurons except the one with the same id.

    This will connect each neuron in each group (default a single group for
    the whole Population) to the target neurons in that group except
    the one with the same id.

    There is also an option to add weights.
    These are then used for each group.

    The know use case is multiple potential winner-takes-all groups where the
    connector each time a neurons spikes will inhibit all other neurons
    in the group.

    As the only know use case is source and target population of the same size
    (including self connections) whose size is an exact positive integer
    multiple of the n_neurons_per_group so that is all currently supported.
    """

    __slots__ = ("__n_neurons_per_group", "__weights")

    def __init__(self, n_neurons_per_group: Optional[int] = None,
                 weights: Optional[NDArray[numpy.float64]] = None,
                 safe: bool = True, verbose: bool = False,
                 callback: None = None):
        """
        :param n_neurons_per_group:
            The number of neurons in each winner-takes-all group.
            Must be a positive integer divisor of source.size
        :param weights:
            The weights for one group of neurons
            Single Value, RandomDistribution and string values not supported.
        :param safe:
            If ``True``, check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.

            .. note::
                Not supported by sPyNNaker.
        """
        super().__init__(safe, callback, verbose)
        self.__n_neurons_per_group = n_neurons_per_group
        self.__weights = weights
        self.__check_weights(weights, n_neurons_per_group)

    def __check_weights(self, weights: Optional[NDArray[numpy.float64]],
                        n_neurons_per_group: Optional[int]) -> None:
        if weights is not None and n_neurons_per_group is not None:
            n_weights = n_neurons_per_group * (n_neurons_per_group - 1)
            if len(weights) != n_weights:
                raise ValueError(
                    "The number of weights must be equal to the number of "
                    f"connections in a group "
                    f"({n_neurons_per_group} x ({n_neurons_per_group} - 1) = "
                    f"{n_weights})")

    def __n_connections(self, synapse_info: SynapseInformation) -> int:
        # If not specified, use the smallest of the two populations
        if self.__n_neurons_per_group is None:
            n_values = min(synapse_info.n_pre_neurons,
                           synapse_info.n_post_neurons)
            return n_values * (n_values - 1)

        # Find out how many groups there are at most
        n_groups_pre = synapse_info.n_pre_neurons // self.__n_neurons_per_group
        n_groups_post = (synapse_info.n_post_neurons //
                         self.__n_neurons_per_group)
        n_groups = min(n_groups_pre, n_groups_post)
        return (n_groups * self.__n_neurons_per_group *
                (self.__n_neurons_per_group - 1))

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_maximum(
            synapse_info.delays, self.__n_connections(synapse_info),
            synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_minimum(
            synapse_info.delays, self.__n_connections(synapse_info),
            synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:

        # At most, a pre-neuron will target all post-neurons in the group,
        # except the one with the same index.  For a given subset of post
        # atoms, there might be fewer to target...
        n_targets = n_post_atoms
        if self.__n_neurons_per_group is not None:
            n_targets = min(self.__n_neurons_per_group - 1, n_post_atoms)
        if min_delay is None or max_delay is None:
            return n_targets

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays, self.__n_connections(synapse_info),
            n_targets, min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        # At most, each post-neuron will be targeted by all pre-neurons in a
        # group, except the one with the same index.
        if self.__n_neurons_per_group is None:
            return min(synapse_info.n_pre_neurons,
                       synapse_info.n_post_neurons) - 1
        return self.__n_neurons_per_group - 1

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        if self.__weights is not None:
            return float(numpy.amax(self.__weights))
        return self._get_weight_maximum(
            synapse_info.weights, self.__n_connections(synapse_info),
            synapse_info)

    @overrides(AbstractConnector.get_weight_mean)
    def get_weight_mean(self, weights: WEIGHTS,
                        synapse_info: SynapseInformation) -> float:
        if self.__weights is None:
            return AbstractConnector.get_weight_mean(
                self, weights, synapse_info)
        else:
            return float(numpy.mean(numpy.abs(self.__weights)))

    @overrides(AbstractConnector.get_weight_variance)
    def get_weight_variance(self, weights: WEIGHTS,
                            synapse_info: SynapseInformation) -> float:
        if self.__weights is None:
            return AbstractConnector.get_weight_variance(
                self, weights, synapse_info)
        else:
            return float(numpy.var(numpy.abs(self.__weights)))

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        group_size = self.__n_neurons_per_group
        if group_size is None:
            group_size = min(synapse_info.n_pre_neurons,
                             synapse_info.n_post_neurons)
        # pylint: disable=protected-access
        post_lo, post_hi = synapse_info.pre_population._view_range
        pre_lo, pre_hi = synapse_info.pre_population._view_range
        post_start = max(post_vertex_slice.lo_atom, post_lo)
        post_end = min(post_vertex_slice.hi_atom + 1, post_hi + 1)
        post_group, post_value = divmod(post_start, group_size)

        pre_start = pre_lo + (post_group * group_size)
        pre_end = min(pre_start + group_size, pre_hi + 1)
        n_values = pre_end - pre_start

        pres = list()
        posts = list()
        for post in range(post_start, post_end):
            for value in range(n_values):
                if value != post_value:
                    pres.append(pre_start + value)
                    posts.append(post)

            post_value += 1
            if post_value == group_size:
                post_value = 0
                pre_start += group_size
                pre_end = min(pre_start + group_size, pre_hi + 1)
                if pre_start >= pre_hi:
                    break
                n_values = pre_end - pre_start
        block = numpy.zeros(len(pres), dtype=self.NUMPY_SYNAPSES_DTYPE)
        block["source"] = pres
        block["target"] = posts
        block["weight"] = self._generate_weights(
            block["source"], block["target"], len(pres), post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], len(pres), post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def __repr__(self) -> str:
        return f"WTAConnector(n_neuron_per_group={self.__n_neurons_per_group})"

    @property
    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_id)
    def gen_connector_id(self) -> int:
        return ConnectorIDs.ALL_BUT_ME_CONNECTOR.value

    @overrides(AbstractGenerateConnectorOnMachine.gen_connector_params)
    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray[uint32]:
        n_values = self.__n_neurons_per_group
        if n_values is None:
            n_values = min(synapse_info.n_pre_neurons,
                           synapse_info.n_post_neurons)
        has_weights = int(self.__weights is not None)
        params = numpy.array([n_values, has_weights], dtype=uint32)
        weights: NDArray[uint32]
        if self.__weights is None:
            weights = numpy.zeros(0, dtype=uint32)
        else:
            weights = DataType.S1615.encode_as_numpy_int_array(self.__weights)
        return numpy.concatenate((params, weights))

    @property
    @overrides(
        AbstractGenerateConnectorOnMachine.gen_connector_params_size_in_bytes)
    def gen_connector_params_size_in_bytes(self) -> int:
        size = BYTES_PER_WORD * 2
        if self.__weights is not None:
            size += len(self.__weights) * BYTES_PER_WORD
        return size

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        if (synapse_info.pre_population.size !=
                synapse_info.post_population.size):
            # Probably works by implementation but there is no know need
            raise NotImplementedError(
                "AllButMeConnector is only designed to be used with "
                "populations that are the same size as each other")
        if self.__n_neurons_per_group is not None:
            if self.__n_neurons_per_group > synapse_info.pre_population.size:
                raise ValueError(
                    "AllButMeConnector cannot be used with a group size "
                    "larger than the population size")
            if ((synapse_info.post_population.size /
                 self.__n_neurons_per_group) !=
                    (synapse_info.post_population.size //
                     self.__n_neurons_per_group)):
                # Probably works by implementation but there is no know need
                raise NotImplementedError(
                    "The number of neurons in each population must be "
                    "divisible by the number of neurons per group")
        n_neurons_per_group = self.__n_neurons_per_group
        if n_neurons_per_group is None:
            n_neurons_per_group = min(synapse_info.pre_population.size,
                                      synapse_info.post_population.size)
        self.__check_weights(self.__weights, n_neurons_per_group)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Tuple
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from pacman.model.graphs.common import Slice


class HasShapeKeyFields(object, metaclass=AbstractBase):
    """
    Indicates a source that has keys in fields for each dimension of the
    source.
    """

    __slots__ = ()

    @abstractmethod
    def get_shape_key_fields(
            self, vertex_slice: Slice) -> List[Tuple[int, int, int, int]]:
        """
        :param vertex_slice:
            The slice of the source vertex
        :returns:
           Get the fields to be used for each dimension in the shape of the
           given source vertex slice, as a list of start, size, mask, shift
           values in the order of the fields.
        """
        raise NotImplementedError



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from contextlib import contextmanager
import logging
import os
from typing import Dict, Iterator, Tuple
import numpy

from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neural_projections import ProjectionApplicationEdge
from spynnaker.pyNN.models.neural_projections import SynapseInformation
from spynnaker.pyNN.models.neuron import ConnectionHolder

logger = FormatAdapter(logging.getLogger(__name__))
_DIRNAME = "synaptic_matrix_reports"


@contextmanager
def _print_all() -> Iterator[None]:
    """
    Update the numpy print options to display everything.
    """
    print_opts = numpy.get_printoptions()
    numpy.set_printoptions(threshold=sys.maxsize)
    try:
        yield
    finally:
        numpy.set_printoptions(**print_opts)


class SpYNNakerSynapticMatrixReport(object):
    """
    Generate the synaptic matrices for reporting purposes.
    """

    def __call__(self, connection_holder: Dict[Tuple[
            ProjectionApplicationEdge, SynapseInformation], ConnectionHolder]
                 ) -> None:
        """
        Convert synaptic matrix for every application edge.

        :param connection_holder: where the synaptic matrices are stored
            (possibly after retrieval from the machine)
        """
        # generate folder for synaptic reports
        top_level_folder = os.path.join(
            SpynnakerDataView.get_run_dir_path(), _DIRNAME)
        if not os.path.exists(top_level_folder):
            os.mkdir(top_level_folder)
        # create progress bar
        progress = ProgressBar(connection_holder.keys(),
                               "Generating synaptic matrix reports")

        # Update the print options to display everything
        with _print_all():
            # for each application edge, write matrix in new file
            for edge, _ in progress.over(connection_holder.keys()):
                # only write matrix's for edges which have matrix's
                if isinstance(edge, ProjectionApplicationEdge):
                    # figure new file name
                    self._write_file(os.path.join(
                        top_level_folder,
                        f"synaptic_matrix_for_application_edge_{edge.label}"),
                        connection_holder, edge)

    def _write_file(
            self, file_name: str, connection_holder:  Dict[Tuple[
                ProjectionApplicationEdge, SynapseInformation],
                ConnectionHolder],
            edge: ProjectionApplicationEdge) -> None:
        # open writer
        try:
            with open(file_name, "w", encoding="utf-8") as f:
                # write all data for all synapse_information's in same file
                for info in edge.synapse_information:
                    f.write(f"{connection_holder[edge, info]}")
        except IOError:
            logger.exception(
                "Generate_placement_reports: Can't open file {} for writing.",
                file_name)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import pyNN.random
from pyNN.random import available_distributions, NumpyRNG
# This file is to work around a Sphinx bug

if "exponential_clipped" not in available_distributions:
    available_distributions["exponential_clipped"] = ('beta', 'low', 'high')


class RandomDistribution(pyNN.random.RandomDistribution):
    """
    Class which defines a next(n) method which returns an array of ``n``
    random numbers from a given distribution.

    Examples::

        >>> rng = RandomDistribution('uniform', (-70, -50))
        >>> rng = RandomDistribution('normal', mu=0.5, sigma=0.1)
        >>> rng = NumpyRNG(seed=8658764)
        >>> rng = RandomDistribution('gamma', k=2.0, theta=5.0, rng=rng)

    .. list-table:: Available distributions
        :widths: auto
        :header-rows: 1

        * - Name
          - Parameters
          - Comments
        * - ``binomial``
          - ``n``, ``p``
          -
        * - ``gamma``
          - ``k``, ``theta``
          -
        * - ``exponential``
          - ``beta``
          -
        * - ``lognormal``
          - ``mu``, ``sigma``
          -
        * - ``normal``
          - ``mu``, ``sigma``
          -
        * - ``normal_clipped``
          - ``mu``, ``sigma``, ``low``, ``high``
          - Values outside (``low``, ``high``) are redrawn
        * - ``normal_clipped_to_boundary``
          - ``mu``, ``sigma``, ``low``, ``high``
          - Values below/above ``low``/``high`` are set to ``low``/``high``
        * - ``poisson``
          - ``lambda_``
          - Trailing underscore since ``lambda`` is a Python keyword
        * - ``uniform``
          - ``low``, ``high``
          -
        * - ``uniform_int``
          - ``low``, ``high``
          - Only generates integer values
        * - ``vonmises``
          - ``mu``, ``kappa``
          -
    """
    # Pylint is wrong about the super-delegation being useless
    def __init__(  # pylint: disable=useless-super-delegation
            self, distribution: str, parameters_pos: Optional[tuple] = None,
            rng: Optional[NumpyRNG] = None, **parameters_named: float):
        """
        :param distribution: the name of a random number distribution.
        :param parameters_pos:
            parameters of the distribution, provided as a tuple. For the
            correct ordering, see `random.available_distributions`.
        :param rng: the random number generator to use, if a specific one is
            desired (e.g., to provide a seed).
        :param parameters_named:
            parameters of the distribution, provided as keyword arguments.

        Parameters may be provided either through ``parameters_pos`` or through
        ``parameters_named``, but not both. All parameters must be provided,
        there are no default values. Parameter names are, in general, as used
        in Wikipedia.
        """
        super().__init__(distribution, parameters_pos, rng, **parameters_named)

    def __repr__(self) -> str:
        return self.__str__()



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Collection, Container, Iterable, List, Optional, Tuple

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged.abstract_sized import Selector

from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice
from pacman.model.routing_info import RoutingInfo
from pacman.utilities.utility_calls import get_keys

from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.abstract_models import HasCustomAtomKeyMap
from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.models.abstract_models import ColouredApplicationVertex
from spynnaker.pyNN.models.common.parameter_holder import ParameterHolder
from spynnaker.pyNN.utilities.buffer_data_type import BufferDataType
from spynnaker.pyNN.models.current_sources import AbstractCurrentSource
from .types import Names, Values


# pylint: disable=abstract-method
class PopulationApplicationVertex(
        ColouredApplicationVertex, HasCustomAtomKeyMap):
    """
    A vertex that can be used in a Population.

    Provides some default functions that can be overridden if the vertex
    supports these.
    """

    # No data required here; makes it easy to mix in!
    __slots__ = ()

    @staticmethod
    def _as_list(names: Names) -> List[str]:
        """
        Normalise the input to a list.

        :param names: The item or items to normalise
        """
        if isinstance(names, str):
            return [names]
        return list(names)

    @staticmethod
    def _check_names(names: Names, allowed: Container[str],
                     type_of_thing: str) -> None:
        """
        Check the list of names are allowed or not.

        :param names: The names to check
        :param allowed: The set of allowed names
        :param type_of_thing: The "type of thing" to report in any error
        :raises KeyError: If one of the names is not allowed
        """
        for name in PopulationApplicationVertex._as_list(names):
            if name not in allowed:
                raise KeyError(f"{name} is not a recognized {type_of_thing}")

    @staticmethod
    def _check_parameters(names: Names, allowed: Container[str]) -> None:
        """
        Check that parameters are allowed.

        :param names: The names to check
        :param allowed: The set of allowed names
        """
        PopulationApplicationVertex._check_names(names, allowed, "parameter")

    @staticmethod
    def _check_variables(names: Names, allowed: Container[str]) -> None:
        """
        Check that state variables are allowed.

        :param names: The names to check
        :param allowed: The set of allowed names
        """
        PopulationApplicationVertex._check_names(names, allowed, "variable")

    def get_parameter_values(
            self, names: Names, selector: Selector = None) -> ParameterHolder:
        """
        :param names: The name or names of the parameter to get
        :param selector: a description of the subrange to accept, or ``None``
            for all. See:
            :py:meth:`~spinn_utilities.ranged.AbstractSized.selector_to_ids`
        :returns: The values of a parameter or parameters for the whole
            Population or a subset if the selector is used.
        :raise KeyError: if the parameter is not something that can be read
        """
        raise KeyError(
            "This Population does not support the reading of parameters")

    def set_parameter_values(
            self, name: str, value: Values, selector: Selector = None) -> None:
        """
        Set the values of a parameter for the whole Population or a subset
        if the selector is used.

        :param name: The name of the parameter to set
        :param value: The new value
        :param selector: a description of the subrange to accept, or ``None``
            for all. See:
            :py:meth:`~spinn_utilities.ranged.AbstractSized.selector_to_ids`
        :raise KeyError: if the parameter is not something that can be changed
        """
        raise KeyError(
            "This Population does not support the setting of parameters")

    def get_parameters(self) -> List[str]:
        """
        :returns: The names of all the parameters that can be obtained
        """
        return []

    def get_initial_state_values(
            self, names: Names, selector: Selector = None) -> ParameterHolder:
        """
        :param names: The name or names of the variable to get
        :param selector: a description of the subrange to accept, or ``None``
            for all. See:
            :py:meth:`~spinn_utilities.ranged.AbstractSized.selector_to_ids`
        :returns: The initial values of a state variable
            for the whole Population
            or a subset if the selector is used.
        :raise KeyError: if the variable is not something that can be read
        """
        raise KeyError(
            "This Population does not support the reading of state"
            " variables")

    def set_initial_state_values(
            self, name: str, value: Values, selector: Selector = None) -> None:
        """
        Set the initial values of a state variable for the whole Population
        or a subset if the selector is used.

        :param name: The name of the variable to set
        :param value: The new value
        :param selector: a description of the subrange to accept, or ``None``
            for all. See:
            :py:meth:`~spinn_utilities.ranged.AbstractSized.selector_to_ids`
        :raise KeyError: if the variable is not something that can be changed
        """
        raise KeyError(
            "This Population does not support the initialization of state"
            " variables")

    def get_current_state_values(
            self, names: Names, selector: Selector = None) -> ParameterHolder:
        """
        :param names: The name or names of the variable to get
        :param selector: a description of the subrange to accept, or ``None``
            for all. See:
            :py:meth:`~spinn_utilities.ranged.AbstractSized.selector_to_ids
        :returns: The current values of a state variable for
            the whole Population or a subset if the selector is used.
        :raise KeyError: if the variable is not something that can be read
        """
        raise KeyError(
            "This Population does not support the reading of state"
            " variables")

    def set_current_state_values(
            self, name: str, value: Values, selector: Selector = None) -> None:
        """
        Set the current values of a state variable for the whole Population
        or a subset if the selector is used.

        :param name: The name of the variable to set
        :param value: The new value
        :param selector: a description of the subrange to accept, or ``None``
            for all. See:
            :py:meth:`~spinn_utilities.ranged.AbstractSized.selector_to_ids`
        :raise KeyError: if the variable is not something that can be changed
        """
        raise KeyError(
            "This Population does not support the setting of state"
            " variables")

    def get_state_variables(self) -> List[str]:
        """
        :returns: A list of supported state variables.
        """
        return []

    def get_units(self, name: str) -> str:
        """
        Get the units of the given parameter or state variable.

        :param name: the name of the parameter to get the units of
        :returns: The unit or this variable. For example 'mV' or 'uS'.
           Will be an empty string for things like spikes and probability
        :raise KeyError:
            if the name isn't recognised or the units cannot be identified
        """
        raise KeyError(f"The units for {name} cannot be found")

    @property
    def conductance_based(self) -> bool:
        """
        Whether the vertex models post-synaptic inputs as
        currents or conductance.

        By default this is False; override if the model accepts
        conductance based input.
        """
        return False

    # recording methods
    # If get_recordable_variables implemented other recording methods must
    # be too

    def get_recordable_variables(self) -> List[str]:
        """
        Get a list of the names of things that can be recorded.

        This methods list the variable recorded via the Population.

        Includes both neuron and synapse recordable things

        :returns: List of variable names
        """
        return []

    def set_recording(
            self, name: str, sampling_interval: Optional[float] = None,
            indices: Optional[Collection[int]] = None) -> None:
        """
        Set a variable recording.

        :param name: The name of the variable to set the status of
        :param sampling_interval:
            How often the variable should be recorded or `None` for every
            time step, in milliseconds
        :param indices: The list of neuron indices to record or `None` for all
        :raises KeyError: if the variable cannot be recorded
        """
        _ = (name, sampling_interval, indices)
        if not self.get_recordable_variables():
            raise KeyError("This Population does not support recording")
        raise NotImplementedError(
            f"{type(self)} has recording variables so should implement "
            f"set_recording")

    def set_not_recording(self, name: str,
                          indices: Optional[Collection[int]] = None) -> None:
        """
        Set a variable not recording.

        :param name: The name of the variable to not record
        :param indices:
            The list of neuron indices to not record or `None` for all
        :raises KeyError: if the variable cannot be stopped from recording
        """
        _ = (name, indices)
        if not self.get_recordable_variables():
            raise KeyError("This Population does not support recording")
        raise NotImplementedError(
            f"{type(self)} has recording variables so should implement "
            f"set_not_recording")

    def get_recording_variables(self) -> List[str]:
        """
        :returns: A list of variables that are currently being recorded.
        """
        if not self.get_recordable_variables():
            return []
        raise NotImplementedError(
            f"{type(self)} has recording variables so should implement "
            f"get_recording_variables")

    def get_buffer_data_type(self, name: str) -> BufferDataType:
        """
        Get the type of data recorded by the buffer manager.

        The buffer data type controls how data returned by the cores is
        handled in NeoBufferDatabase.

        :param name: The name of the variable recorded
        :raises KeyError: if the variable isn't being recorded
        :returns: Data type in the Buffer Database
        """
        if name not in self.get_recordable_variables():
            raise KeyError(f"{name} is not being recorded")
        raise NotImplementedError(
            f"{type(self)} has recording variables so should implement "
            f"get_recording_variables")

    def get_sampling_interval_ms(self, name: str) -> float:
        """
        Get the sampling interval of the recording for the given variable.

        The values is in ms and unless selective recording is used will be
        `SpynnakerDataView.get_simulation_time_step_us()`

        :param name: The name of the variable
        :returns: Sampling rate in milliseconds for this variable
        :raises KeyError: If the variable isn't being recorded
        """
        if name not in self.get_recordable_variables():
            raise KeyError(f"{name} is not being recorded")
        raise NotImplementedError(
            f"{type(self)} has recording variables so should implement "
            f"get_recording_variables")

    def get_data_type(self, name: str) -> Optional[DataType]:
        """
        Get the type data returned by a recording of the variable.

        This is the type of data the C code is returning.
        For instance data such as spikes this will be `None`.

        :param name: The name of the variable to get the type of
        :returns: The data type used by the C code
        :raise KeyError: If the variable isn't recordable
        """
        if name not in self.get_recordable_variables():
            raise KeyError(f"{name} is not being recorded")
        raise NotImplementedError(
            f"{type(self)} has recording variables so should implement "
            f"get_data_type")

    def get_recording_region(self, name: str) -> int:
        """
        :param name: The name of the variable to get the region of
        :returns: The recording region for the named variable.
        :raises KeyError: If the variable isn't being recorded
        """
        if name not in self.get_recordable_variables():
            raise KeyError(f"{name} is not being recorded")
        raise NotImplementedError(
            f"{type(self)} has recording variables so should implement "
            f"get_recording_region")

    def get_neurons_recording(
            self, name: str, vertex_slice: Slice) -> Optional[Collection[int]]:
        """
        Gets the neurons being recorded on the core with this slice.

        Typically `vertex_slice.get_raster_ids(atoms_shape)`
        but may be a sublist if doing selective recording.

        :param name: The name of the variable to get the region of
        :param vertex_slice:
        :return: A list of the global raster IDs of the atoms in recording
            named variable within this slice
        """
        _ = (name, vertex_slice)
        if name not in self.get_recordable_variables():
            raise KeyError(f"{name} is not being recorded")
        raise NotImplementedError(
            f"{type(self)} has recording variables so should implement "
            f"get_neurons_recording")

    # end of recording methods

    def inject(self, current_source: AbstractCurrentSource,
               selector: Selector = None) -> None:
        """
        Inject a current source into this population.

        :param current_source: the current source to be injected
        :param selector: a description of the subrange to accept, or ``None``
            for all. See:
            :py:meth:`~spinn_utilities.ranged.AbstractSized.selector_to_ids`
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            if the population doesn't support injection
        """
        raise ConfigurationException(
            "This Population doesn't support injection")

    @property
    @overrides(ColouredApplicationVertex.n_colour_bits)
    def n_colour_bits(self) -> int:
        return 0

    @overrides(HasCustomAtomKeyMap.get_atom_key_map)
    def get_atom_key_map(
            self, pre_vertex: MachineVertex, partition_id: str,
            routing_info: RoutingInfo) -> Iterable[Tuple[int, int]]:
        base_key = routing_info.get_key_from(
            pre_vertex, partition_id)
        vertex_slice = pre_vertex.vertex_slice
        keys = get_keys(base_key, vertex_slice, self.n_colour_bits)
        return zip(vertex_slice.get_raster_ids(), keys)



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from collections import defaultdict
import csv
from datetime import datetime
import logging
from typing import (
    Any, Dict, Iterable, List, Optional, Tuple, Union, TYPE_CHECKING)

from neo import AnalogSignal, Block, Event, Segment, SpikeTrain
import numpy
from numpy import integer, float64
from numpy.typing import NDArray
from quantities import Quantity, ms

from spinn_utilities.log import FormatAdapter

from spynnaker.pyNN.data import SpynnakerDataView

if TYPE_CHECKING:
    from _csv import _writer as CSVWriter, _reader as CSVReader
    from spynnaker.pyNN.utilities.neo_buffer_database import Annotations

logger = FormatAdapter(logging.getLogger(__name__))


class NeoCsv(object):
    """
    Code to read a csv file and create a neo object.

    """
    _POPULATION = "population"
    _DESCRIPTION = "description"
    _SIZE = "size"
    _FIRST_ID = "first_id"
    _SIMULATOR = "simulator"
    _DT = "dt"  # t_stop

    _INDEXES = "indexes"
    _NO_INTERSECTION = "no intersection between recording and view indexes"

    _SEGMENT_NUMBER = 'segment_number'
    _REC_DATETIME = "rec_datetime"

    _T_START = "t_start"
    _T_STOP = "t_stop"
    _SAMPLING_PERIOD = "sampling_period"
    _UNITS = "units"

    _MATRIX = "matrix"

    _EVENT = "event"
    _ELMINATION = "elimination"
    _FORMATION = "formation"

    _SPIKES = "spikes"

    def _csv_variable_metdata(
            self, csv_writer: CSVWriter, variable_type: str, variable: str,
            t_start: float, t_stop: float, sampling_interval_ms: float,
            units: Optional[str]) -> None:
        """
        Writes the metadata for a variable to CSV

        :param csv_writer: Open CSV writer to write to
        :param variable_type:
        :param variable:
        :param t_start:
        :param t_stop:
        :param sampling_interval_ms:
        :param units:
        """
        csv_writer.writerow([variable_type, variable])
        csv_writer.writerow([self._T_START, t_start * ms])
        csv_writer.writerow([self._T_STOP, t_stop * ms])
        sampling_period = sampling_interval_ms * ms
        csv_writer.writerow([self._SAMPLING_PERIOD, sampling_period])
        if units is None:
            units = "dimensionless"
        csv_writer.writerow([self._UNITS, units])
        csv_writer.writerow([])

    def __quantify(self, as_str: str) -> Quantity:
        """
        Converts a String into a quantities.Quantity

        The String should be a float, a space and a Quantities label

        :param as_str: String representation of a quantity.
        :return: A Quantities object
        """
        parts = as_str.split(" ")
        return Quantity(float(parts[0]), units=parts[1])

    def __read_variable_metadata(self, csv_reader: CSVReader) -> Tuple[
            Quantity, Quantity, Quantity, str]:
        """
        Reads a block of metadata, formats it and returns it as a dict

        A block is a number of rows each of exactly 2 values followed by an
        empty row

        :param csv_reader: Open CSV writer to read from
        :return: t_start, t_stop, sampling_period, units
        """
        metadata = self.__read_metadata(csv_reader)
        return (
            self.__quantify(metadata[self._T_START]),
            self.__quantify(metadata[self._T_STOP]),
            self.__quantify(metadata[self._SAMPLING_PERIOD]),
            metadata[self._UNITS])

    def __read_signal_array(self, csv_reader: CSVReader) -> NDArray[float64]:
        """
        Reads a block of data and converts it in a numpy array.

        A block is a number of rows followed by an empty row.
        All rows must have the same length.
        The assumption is that all values in the block represent floats.

        :param csv_reader: Open CSV reader to read from
        :return: Numpy signal array of floats
        """
        rows = []
        row = next(csv_reader)
        while len(row) > 0:
            rows.append(row)
            row = next(csv_reader)
        return numpy.asarray(rows, dtype=float64)

    def __csv_indexes(self, indexes: NDArray[integer],
                      csv_writer: CSVWriter) -> None:
        """
        Writes the indexes for which there could be data to the CSV.

        The indexes will be the intersection of the view used to record
        and the view used to write the CSV.

        If there is no indexes this will write the NO_INTERESTION string

        There may be a case where this writes indexes and then there is still
        no data below. This happens when the data is an empty array
        such as no spikes or rewires happened or the run was for time 0

        :param csv_writer: Open CSV writer to write to
        :param indexes:
        """
        if len(indexes) > 0:
            csv_writer.writerow(indexes)
        else:
            csv_writer.writerow([self._NO_INTERSECTION])

    def __read_indexes(self, csv_reader: CSVReader) -> NDArray[integer]:
        """
        Reads the index or NO_INTERSECTION string from the CSV

        :param csv_reader: Open CSVreader to read from
        :return: list of indexes
        """
        row = next(csv_reader)
        assert len(row) > 0
        if len(row) == 1:
            if row[0] == self._NO_INTERSECTION:
                return numpy.array([], dtype=int)
        return numpy.asarray(row, dtype=int)

    def _insert_spike_data(
            self, view_indexes: Iterable[int], segment: Segment,
            spikes: NDArray, t_start: float, t_stop: float,
            sampling_rate: Quantity) -> None:
        """
        Creates the SpikeTrains and inserts then into the segment

        :param view_indexes:
        :param segment:
        :param spikes:
        :param t_start:
        :param t_stop:
        :param sampling_rate: Rate a neuron is recorded
        """
        block = segment.block
        first_id = block.annotations[self._FIRST_ID]
        times = defaultdict(list)
        for neuron_id, time in spikes:
            times[int(neuron_id)].append(time)

        for index in view_indexes:
            spiketrain = SpikeTrain(
                times=times[index],
                t_start=t_start,
                t_stop=t_stop,
                units=ms,
                dtype=float64,
                sampling_rate=sampling_rate,
                source_population=block.name,
                source_id=index + first_id,
                source_index=index)
            segment.spiketrains.append(spiketrain)

    def _csv_spike_data(self, csv_writer: CSVWriter, spikes: NDArray,
                        indexes: NDArray[integer]) -> None:
        """
        Writes the spikes to the CSV file.

        :param csv_writer: Open CSV writer to write to
        :param spikes:
        :param indexes: The indexes for which there could be data
        """
        self.__csv_indexes(indexes, csv_writer)
        csv_writer.writerows(spikes)
        csv_writer.writerow([])

    def __read_spike_data(self, csv_reader: CSVReader, segment: Segment,
                          variable: str) -> None:
        """
        Reads spikes from the CSV file and add SpikeTrains to the segment.

        :param csv_reader: Open CSV writer to read from
        :param segment:
        :param variable: Name of the variable being read
        """
        try:
            t_start, t_stop, sampling_period, _ = \
                self.__read_variable_metadata(csv_reader)
            indexes = self.__read_indexes(csv_reader)
            spikes = self.__read_signal_array(csv_reader)
            sampling_rate = 1000 / sampling_period
            self._insert_spike_data(
                indexes, segment, spikes, t_start, t_stop, sampling_rate)
        except KeyError as ex:
            logger.exception("Metadata for {} is missing {}. "
                             "So this data will be skipped", variable, ex)
            return

    def _insert_matrix_data(
            self, variable: str, segment: Segment, signal_array: NDArray,
            indexes: NDArray[integer], t_start: float, sampling_rate: Quantity,
            units: Union[Quantity, str, None]) -> None:
        """
        Adds a data item that is an analog signal to a neo segment.

        :param variable: the variable name
        :param segment: Segment to add data to
        :param signal_array: the raw signal data
        :param indexes: The indexes for the data
        :param sampling_rate: Rate a neuron is recorded
        :param units: the units of the recorded value
        """
        block = segment.block

        first_id: int = block.annotations[self._FIRST_ID]

        ids = list(indexes + first_id)
        if units is None:
            units = "dimensionless"
        data_array = AnalogSignal(
            signal_array,
            units=units,
            t_start=t_start,
            sampling_rate=sampling_rate,
            name=variable,
            source_population=block.name,
            source_ids=ids,
            channel_names=indexes)
        data_array.shape = (data_array.shape[0], data_array.shape[1])
        segment.analogsignals.append(data_array)

    def _csv_matrix_data(
            self, csv_writer: CSVWriter, signal_array: NDArray,
            indexes: NDArray[integer]) -> None:
        """
        Writes data to a CSV file.

        :param csv_writer: Open CSV writer to write to
        :param signal_array: the raw signal data
        :param indexes: The indexes for the data
        """
        self.__csv_indexes(indexes, csv_writer)
        csv_writer.writerows(signal_array)
        csv_writer.writerow([])

    def __read_matrix_data(self, csv_reader: CSVReader, segment: Segment,
                           variable: str) -> None:
        """
        Reads matrix data and adds it to the segment.

        :param csv_reader: Open CSV writer to read from
        :param segment:
        :param variable:
        """
        t_start, _, sampling_period, units = \
            self.__read_variable_metadata(csv_reader)
        indexes = self.__read_indexes(csv_reader)
        signal_array = self.__read_signal_array(csv_reader)
        sampling_rate = 1000 / sampling_period
        self._insert_matrix_data(
            variable, segment, signal_array, indexes,
            t_start, sampling_rate, units)

    def _insert_formation_events(
            self, segment: Segment, variable: str,
            formation_times: List[Quantity],
            formation_labels: List[str]) -> None:
        """
        Adds formation data to a neo segment.

        :param segment: Segment to add data to
        :param variable: the variable name
        :param formation_times:
        :param formation_labels:
        """
        formation_event_array = Event(
            times=formation_times,
            labels=formation_labels,
            units="ms",
            name=variable + "_form",
            description="Synapse formation events",
            array_annotations={})
        segment.events.append(formation_event_array)

    def _insert_elimination_events(
            self, segment: Segment, variable: str,
            elimination_times: List[Quantity],
            elimination_labels: List[str]) -> None:
        """
        Adds elimination data to a neo segment.

        :param segment: Segment to add data to
        :param variable: the variable name
        :param elimination_times:
        :param elimination_labels:
        """
        elimination_event_array = Event(
            times=elimination_times,
            labels=elimination_labels,
            units="ms",
            name=variable + "_elim",
            description="Synapse elimination events",
            array_annotations={})
        segment.events.append(elimination_event_array)

    def _insert_neo_rewirings(self, segment: Segment, event_array: NDArray,
                              variable: str) -> None:
        """
        Adds data that represent rewire events to a neo segment.

        :param segment: Segment to add data to
        :param event_array: the raw "event" data
        :param variable: the variable name
        """
        formation_times: List[Quantity] = []
        formation_labels: List[str] = []
        elimination_times: List[Quantity] = []
        elimination_labels: List[str] = []

        for event in event_array:
            event_time = event[0] * ms
            pre_id = int(event[1])
            post_id = int(event[2])
            if event[3] == 1:
                formation_times.append(event_time)
                formation_labels.append(f"{pre_id}_{post_id}_formation")
            else:
                elimination_times.append(event_time)
                elimination_labels.append(f"{pre_id}_{post_id}_elimination")

        self._insert_formation_events(
            segment, variable, formation_times, formation_labels)
        self._insert_elimination_events(
            segment, variable, elimination_times, elimination_labels)

    def _csv_rewirings(
            self, csv_writer: CSVWriter, event_array: NDArray) -> None:
        """
        Adds data that represent rewires events to a CSV file.

        :param csv_writer: Open CSV writer to write to
        :param event_array: the raw "event" data
        """
        formation: List[Tuple[Quantity, str]] = []
        elimination: List[Tuple[Quantity, str]] = []

        for event in event_array:
            event_time = event[0] * ms
            pre_id = int(event[1])
            post_id = int(event[2])
            if event[3] == 1:
                formation.append(
                    (event_time, f"{pre_id}_{post_id}_formation"))
            else:
                elimination.append(
                    (event_time, f"{pre_id}_{post_id}_elimination"))

        csv_writer.writerow([self._FORMATION])
        csv_writer.writerows(formation)
        csv_writer.writerow([])
        csv_writer.writerow([self._ELMINATION])
        csv_writer.writerows(elimination)
        csv_writer.writerow([])

    def __read_times_and_labels(self, csv_reader: CSVReader) -> Tuple[
            List[Quantity], List[str]]:
        """
        Reads formation or elimination data from the CSV file.

        :param csv_reader: Open CSV writer to read from
        :return: A list of times and a list of labels
        """
        times: List[Quantity] = []
        labels: List[str] = []
        row = next(csv_reader)
        while len(row) > 0:
            assert len(row) == 2
            times.append(self.__quantify(row[0]))
            labels.append(row[1])
            row = next(csv_reader)
        return times, labels

    def __read_rewirings(self, csv_reader: CSVReader, segment: Segment,
                         variable: str) -> None:
        """
        Reads rewiring data from a CSV file and adds it to the segment.

        :param csv_reader: Open CSV writer to read from
        :param segment: Segment to add data to
        :param variable:
        """
        self.__read_metadata(csv_reader)
        row = next(csv_reader)
        assert row == ["formation"]
        times, labels = self.__read_times_and_labels(csv_reader)
        self._insert_formation_events(
            segment, variable, times, labels)
        row = next(csv_reader)
        assert row == ["elimination"]
        times, labels = self.__read_times_and_labels(csv_reader)
        self._insert_elimination_events(
            segment, variable, times, labels)

    def _insert_empty_segment(self, block: Block, segment_number: int,
                              rec_datetime: datetime) -> Segment:
        """
        Creates an empty segment and adds it to the block.

        Unless other insert methods are called the segment will hold no data.
        """
        segment = Segment(
            name=f"segment{segment_number}",
            description=block.description,
            rec_datetime=rec_datetime)
        for i in range(len(block.segments), segment_number):
            block.segments.append(Segment(
                name=f"segment{i}",
                description="empty"))
        if segment_number in block.segments:
            block.segments[segment_number] = segment
        else:
            block.segments.append(segment)
        segment.block = block
        if block.rec_datetime is None:
            block.rec_datetime = rec_datetime

        return segment

    def _csv_segment_metadata(self, csv_writer: CSVWriter, segment_number: int,
                              rec_datetime: datetime) -> None:
        """
        Writes only the segment's metadata to CSV.

        Unless other CSV methods are called the CSV will hold no data.

        :param csv_writer: Open CSV writer to write to
        :param segment_number:
        :param rec_datetime:
        """
        csv_writer.writerow([self._SEGMENT_NUMBER, segment_number])
        csv_writer.writerow([self._REC_DATETIME, rec_datetime])
        csv_writer.writerow([])

    def __read_segment(self, csv_reader: CSVReader, block: Block,
                       segment_number_st: str) -> Segment:
        """
        Reads only segments metadata and inserts an empty segment.

        Unless other read methods are called the segment will hold no data

        :param csv_reader: Open CSV writer to read from
        :param block:
        :param segment_number_st:
        """
        row = next(csv_reader)
        assert (row[0] == self._REC_DATETIME)
        rec_datetime = datetime.strptime(row[1], '%Y-%m-%d %H:%M:%S.%f')
        # consume the empty line
        next(csv_reader)
        return self._insert_empty_segment(
            block, int(segment_number_st), rec_datetime)

    def _insert_empty_block(
            self, pop_label: str, description: str, size: int, first_id: int,
            dt: float, simulator: str,
            annotations: Annotations = None) -> Block:
        """
        Creates and empty Neo block object with just metadata.

        Unless other insert methods are called this block will hold no data

        Does not actually "insert" the block anywhere.
        The insert name is aligned with other methods to create a full block

        :return: a block with just metadata
        """
        block = Block()
        block.name = pop_label
        block.description = description
        metadata: Dict[str, Any] = {}
        metadata[self._SIZE] = size
        metadata["first_index"] = 0
        metadata['last_index'] = size,
        metadata[self._FIRST_ID] = first_id
        metadata['last_id'] = first_id + size,
        metadata['label'] = pop_label
        metadata[self._SIMULATOR] = simulator
        metadata[self._DT] = dt
        block.annotate(**metadata)
        if annotations:
            block.annotate(**annotations)
        return block

    def _csv_block_metadata(
            self, csv_writer: CSVWriter, pop_label: str, t_stop: float,
            pop_size: int, first_id: int, description: str,
            annotations: Annotations) -> None:
        """
        :param csv_writer: Open CSV writer to write to
        :param pop_label:
        :param t_stop:
        :param pop_size:
        :param first_id:
        :param description:
        :param annotations: annotations to put on the neo block
        """
        csv_writer.writerow([self._POPULATION, pop_label])
        csv_writer.writerow([self._DESCRIPTION, f'"{description}"'])

        csv_writer.writerow([self._SIZE, pop_size])
        csv_writer.writerow([self._FIRST_ID, first_id])
        csv_writer.writerow(
            [self._SIMULATOR, SpynnakerDataView.get_sim_name()])
        csv_writer.writerow([self._DT, t_stop])
        # does not make sense on Spinnaker but oh well
        if annotations:
            for key, value in annotations.items():
                csv_writer.writerow([str(key), str(value)])
        csv_writer.writerow([])

    def __read_empty_block(self, csv_reader: CSVReader) -> Block:
        """
        Reads block metadata and uses it to create an empty block.

        :param csv_reader: Open CSV writer to read from
        :return: empty block
        """
        metadata = self.__read_metadata(csv_reader)
        return self._insert_empty_block(
            pop_label=metadata.pop(self._POPULATION),
            description=metadata.pop(self._DESCRIPTION),
            size=int(metadata.pop(self._SIZE)),
            first_id=int(metadata.pop(self._FIRST_ID)),
            dt=float(metadata.pop(self._DT)),
            simulator=metadata.pop(self._SIMULATOR),
            annotations=metadata)

    def __read_metadata(self, csv_reader: CSVReader) -> Dict[str, str]:
        """
        Reads a block of metadata and converts it to a dict.

        A metadata block is zero or more lines of two columns followed by an
        empty line. the first column will be the keys the second the data

        :param csv_reader: Open CSV writer to read from
        :return: a dict of the keys to unformatted values
        """
        metadata: Dict[str, str] = {}
        row = next(csv_reader)
        while len(row) > 0:
            assert len(row) == 2
            metadata[row[0]] = row[1]
            row = next(csv_reader)
        return metadata

    def read_csv(self, csv_file: str) -> Block:
        """
        Reads a whole CSV file and creates a block with data.

        :param csv_file: Path of file to read
        :return: a block with all the data in the CSV file.
        """
        with open(csv_file, newline='',  encoding="utf-8") as csvfile:
            csv_reader = csv.reader(csvfile, delimiter=',', quotechar='"')
            block = self.__read_empty_block(csv_reader)
            category = block
            while True:
                try:
                    try:
                        row = next(csv_reader)
                        category = row[0]
                    except IndexError:
                        logger.warning("Ignoring extra blank line after {}",
                                       category)
                        row = next(csv_reader)
                        while len(row) == 0:
                            row = next(csv_reader)
                        category = row[0]

                    if row[0] == self._SEGMENT_NUMBER:
                        segment = self.__read_segment(
                            csv_reader, block, row[1])
                    elif row[0] == self._MATRIX:
                        self.__read_matrix_data(
                            csv_reader, segment, row[1])
                    elif row[0] == self._SPIKES:
                        self.__read_spike_data(
                            csv_reader, segment, row[1])
                    elif row[0] == self._EVENT:
                        self.__read_rewirings(
                            csv_reader, segment, row[1])
                    else:
                        logger.error("ignoring csv block starting with {}",
                                     row[0])
                        # ignore a block
                        row = next(csv_reader)
                        while len(row) > 0:
                            row = next(csv_reader)
                except StopIteration:
                    return block



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .push_bot_device import PushBotEthernetDevice
from .push_bot_laser_device import PushBotEthernetLaserDevice
from .push_bot_led_device import PushBotEthernetLEDDevice
from .push_bot_motor_device import PushBotEthernetMotorDevice
from .push_bot_retina_device import PushBotEthernetRetinaDevice
from .push_bot_speaker_device import PushBotEthernetSpeakerDevice
from .push_bot_retina_connection import PushBotRetinaConnection
from .push_bot_translator import PushBotTranslator
from .push_bot_wifi_connection import (
    get_pushbot_wifi_connection, PushBotWIFIConnection)

__all__ = ["PushBotEthernetDevice", "PushBotEthernetLaserDevice",
           "PushBotEthernetLEDDevice", "PushBotEthernetMotorDevice",
           "PushBotEthernetRetinaDevice", "PushBotEthernetSpeakerDevice",
           "PushBotRetinaConnection", "PushBotTranslator",
           "get_pushbot_wifi_connection", "PushBotWIFIConnection"]



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import math
from typing import Iterable, Optional, TYPE_CHECKING, Tuple

import numpy
from numpy import uint32
from numpy.typing import NDArray

from pacman.model.graphs.application import ApplicationEdge
from pacman.model.partitioner_splitters import AbstractSplitterCommon

from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.data import SpynnakerDataView

if TYPE_CHECKING:
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge)

#: number of elements
#  key, n atoms, atoms_per_core, pointer to bitfield
FILTER_INFO_WORDS = 4

#: n_filters, pointer for array
FILTER_HEADER_WORDS = 2

#: the number of bits in a word
# (WHY IS THIS NOT A CONSTANT SOMEWHERE!)
BIT_IN_A_WORD = 32.0


def is_sdram_poisson_source(app_edge: ApplicationEdge) -> bool:
    """
    :returns: True if a given app edge is a poisson source being sent over
        SDRAM as it can likely be discounted if so
    """
    # Avoid circular import
    # pylint: disable=import-outside-toplevel
    from spynnaker.pyNN.extra_algorithms.splitter_components import (
        SplitterPoissonDelegate)
    splitter: AbstractSplitterCommon = app_edge.pre_vertex.splitter
    if isinstance(splitter, SplitterPoissonDelegate):
        if splitter.send_over_sdram:
            return True
    return False


def _unique_edges(projections: Iterable[Projection]) -> Iterable[
        Tuple[ProjectionApplicationEdge, str]]:
    """
    Get the unique application edges of a collection of projections.

    :param projections: The projections to examine.
    """
    seen_edges = set()
    for proj in projections:
        # pylint: disable=protected-access
        edge = proj._projection_edge
        synapse_info = proj._synapse_information
        if is_sdram_poisson_source(edge):
            continue
        if (edge, synapse_info.partition_id) not in seen_edges:
            seen_edges.add((edge, synapse_info.partition_id))
            yield edge, synapse_info.partition_id


def get_sdram_for_bit_field_region(
        incoming_projections: Iterable[Projection]) -> int:
    """
    The SDRAM for the bit field filter region.

    :param incoming_projections:
        The projections that target the vertex in question
    :return: the estimated number of bytes used by the bit field region
    """
    sdram = FILTER_HEADER_WORDS * BYTES_PER_WORD
    for in_edge, _part_id in _unique_edges(incoming_projections):
        n_atoms = in_edge.pre_vertex.n_atoms
        n_words_for_atoms = int(math.ceil(n_atoms / BIT_IN_A_WORD))
        sdram += (FILTER_INFO_WORDS + n_words_for_atoms) * BYTES_PER_WORD
        # Also add for delay vertices if needed
        n_words_for_delays = int(math.ceil(
            n_atoms * in_edge.n_delay_stages / BIT_IN_A_WORD))
        sdram += (FILTER_INFO_WORDS + n_words_for_delays) * BYTES_PER_WORD
    return sdram


def get_sdram_for_keys(incoming_projections: Iterable[Projection]) -> int:
    """
    Gets the space needed for keys.

    :param incoming_projections:
        The projections that target the vertex in question
    :return: SDRAM needed
    """
    # basic sdram
    sdram = 0
    for in_edge, _part_id in _unique_edges(incoming_projections):
        sdram += BYTES_PER_WORD
        if in_edge.n_delay_stages:
            sdram += BYTES_PER_WORD
    return sdram


def get_bitfield_key_map_data(
        incoming_projections: Iterable[Projection]) -> NDArray[uint32]:
    """
    :param incoming_projections:
        The projections to generate bitfields for
    :returns: Data for the key map region.
    """
    # Gather the source vertices that target this core
    routing_infos = SpynnakerDataView.get_routing_infos()
    sources = []
    for in_edge, part_id in _unique_edges(incoming_projections):
        key = routing_infos.get_key_from(
            in_edge.pre_vertex, part_id)
        sources.append([key, in_edge.pre_vertex.n_atoms])
        if in_edge.delay_edge is not None:
            delay_key = routing_infos.get_key_from(
                in_edge.delay_edge.pre_vertex, part_id)
            n_delay_atoms = (
                in_edge.pre_vertex.n_atoms * in_edge.n_delay_stages)
            sources.append([delay_key, n_delay_atoms])

    if not sources:
        return numpy.array([], dtype=uint32)

    # Make keys and atoms, ordered by keys
    key_map = numpy.array(sources, dtype=uint32)
    key_map = key_map[numpy.argsort(key_map[:, 0])]

    # get the number of atoms per item
    return key_map[:, 1]


def write_bitfield_init_data(
        spec: DataSpecificationBase, bit_field_region: int,
        n_bit_field_bytes: int,
        bit_field_region_ref: Optional[int] = None) -> None:
    """
    Writes the initialisation data needed for the bitfield generator.

    :param spec: data specification writer
    :param bit_field_region: the region ID for the bit-field filters
    :param n_bit_field_bytes: the size of the region
    :param bit_field_region_ref: The reference to the region
    """
    # reserve the final destination for the bitfields
    spec.reserve_memory_region(
        region=bit_field_region, size=n_bit_field_bytes,
        label="bit_field region", reference=bit_field_region_ref)

    # Ensure a 0 is written at least to indicate no bit fields if not expanded
    spec.switch_write_focus(bit_field_region)
    spec.write_value(0)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary

from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.models.neuron.implementations import ModelParameter
from spynnaker.pyNN.utilities.struct import Struct

from .abstract_threshold_type import AbstractThresholdType

V_THRESH = "v_thresh"


class ThresholdTypeStatic(AbstractThresholdType):
    """
    A threshold that is a static value.
    """
    __slots__ = ("__v_thresh", )

    def __init__(self, v_thresh: ModelParameter):
        """
        :param v_thresh: :math:`V_{thresh}`
        """
        super().__init__(
            [Struct([(DataType.S1615, V_THRESH)])],
            {V_THRESH: "mV"})
        self.__v_thresh = v_thresh

    @overrides(AbstractThresholdType.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[V_THRESH] = self._convert(self.__v_thresh)

    @overrides(AbstractThresholdType.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        pass

    @property
    def v_thresh(self) -> ModelParameter:
        """
        :math:`V_{thresh}`
        """
        return self.__v_thresh



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_has_a_plus_a_minus import AbstractHasAPlusAMinus
from .abstract_weight_dependence import AbstractWeightDependence
from .weight_dependence_additive import WeightDependenceAdditive
from .weight_dependence_multiplicative import WeightDependenceMultiplicative
from .weight_dependence_additive_triplet import WeightDependenceAdditiveTriplet

__all__ = ["AbstractHasAPlusAMinus", "AbstractWeightDependence",
           "WeightDependenceAdditive", "WeightDependenceMultiplicative",
           "WeightDependenceAdditiveTriplet"]



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .connection_holder import ConnectionHolder
from .population_machine_vertex import (
    PopulationMachineVertex, SpikeProcessingProvenance)
from .population_neurons_machine_vertex import PopulationNeuronsMachineVertex
from .population_machine_neurons import NeuronProvenance, NeuronRegions
from .population_synapses_machine_vertex_lead import (
    PopulationSynapsesMachineVertexLead)
from .population_synapses_machine_vertex_shared import (
    PopulationSynapsesMachineVertexShared)
from .population_synapses_machine_vertex_common import (
    PopulationSynapsesMachineVertexCommon, SpikeProcessingFastProvenance)
from .population_machine_synapses_provenance import SynapseProvenance
from .population_vertex import PopulationVertex
from .abstract_pynn_neuron_model import AbstractPyNNNeuronModel
from .abstract_pynn_neuron_model_standard import (
    AbstractPyNNNeuronModelStandard)
from .population_machine_local_only_combined_vertex import (
    PopulationMachineLocalOnlyCombinedVertex, LocalOnlyProvenance)

__all__ = ["AbstractPyNNNeuronModel",
           "AbstractPyNNNeuronModelStandard", "ConnectionHolder",
           "PopulationMachineVertex", "PopulationNeuronsMachineVertex",
           "NeuronProvenance", "PopulationSynapsesMachineVertexCommon",
           "PopulationSynapsesMachineVertexLead", "NeuronRegions",
           "PopulationSynapsesMachineVertexShared", "PopulationVertex",
           "SynapseProvenance",
           "SpikeProcessingProvenance", "SpikeProcessingFastProvenance",
           "PopulationMachineLocalOnlyCombinedVertex", "LocalOnlyProvenance"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from decimal import Decimal
from enum import Enum
from typing import Optional
from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class SendType(Enum):
    """
    The data type to be sent in the payload of the multicast packet.
    """
    #: Signed integer
    SEND_TYPE_INT = 0
    #: Unsigned integer
    SEND_TYPE_UINT = 1
    #: Signed accum (s15.16)
    SEND_TYPE_ACCUM = 2
    #: Unsigned accum (u16.16)
    SEND_TYPE_UACCUM = 3
    #: Signed fract (s0.31)
    SEND_TYPE_FRACT = 4
    #: Unsigned fract (u0.32)
    SEND_TYPE_UFRACT = 5


class AbstractMulticastControllableDevice(object, metaclass=AbstractBase):
    """
    A device that can be controlled by sending multicast packets to it,
    either directly, or via Ethernet using an AbstractEthernetTranslator.
    """
    __slots__ = ()

    @property
    @abstractmethod
    def device_control_partition_id(self) -> str:
        """
        A partition ID to give to an outgoing edge partition that will
        control this device.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def device_control_key(self) -> int:
        """
        The key that must be sent to the device to control it.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def device_control_uses_payload(self) -> bool:
        """
        Whether the control of the device accepts an arbitrary valued
        payload, the value of which will change the devices behaviour.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def device_control_min_value(self) -> float:
        """
        The minimum value to send to the device.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def device_control_max_value(self) -> Decimal:
        """
        The maximum value to send to the device.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def device_control_timesteps_between_sending(self) -> Optional[int]:
        """
        The number of timesteps between sending commands to the device.
        This defines the "sampling interval" for the device.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def device_control_send_type(self) -> SendType:
        """
        The type of data to be sent.
        """
        raise NotImplementedError

    @property
    def device_control_scaling_factor(self) -> int:  # pragma: no cover
        """
        The scaling factor used to send the payload to this device.
        """
        return 1

    @property
    def device_control_first_send_timestep(self) -> Optional[int]:
        """
        The first timestep that the device should send in (0 by default).
        """
        return 0



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import List, Iterable, TYPE_CHECKING
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
if TYPE_CHECKING:
    from .abstract_ethernet_translator import AbstractEthernetTranslator
    from .abstract_multicast_controllable_device import (
        AbstractMulticastControllableDevice)


class AbstractEthernetController(object, metaclass=AbstractBase):
    """
    A controller that can send multicast packets which can be received
    over Ethernet and translated to control an external device.
    """
    __slots__ = ()

    @abstractmethod
    def get_message_translator(self) -> AbstractEthernetTranslator:
        """
        :returns: The translator of messages.
        """
        raise NotImplementedError

    @abstractmethod
    def get_external_devices(self) -> Iterable[
            AbstractMulticastControllableDevice]:
        """
        :returns:
           The external devices that are to be controlled by the controller.
        """
        raise NotImplementedError

    @abstractmethod
    def get_outgoing_partition_ids(self) -> List[str]:
        """
        :returns: The partition IDs of messages coming out of the controller.
        """
        raise NotImplementedError



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable

import numpy
from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationBase)
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BYTES_PER_SHORT)

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neuron.plasticity.stdp.synapse_structure import (
    SynapseStructureWeightAccumulator)
from spynnaker.pyNN.models.neuron.plasticity.stdp.common import (
    STDP_FIXED_POINT_ONE)

from .abstract_timing_dependence import AbstractTimingDependence


class TimingDependenceRecurrent(AbstractTimingDependence):
    """
    A timing dependence STDP rule based on recurrences.
    """
    __slots__ = (
        "__accumulator_depression_plus_one",
        "__accumulator_potentiation_minus_one",
        "__dual_fsm",
        "__mean_post_window",
        "__mean_pre_window",
        "__a_plus",
        "__a_minus")

    __PARAM_NAMES = (
        'accumulator_depression', 'accumulator_potentiation',
        'mean_pre_window', 'mean_post_window', 'dual_fsm')

    def __init__(
            self, accumulator_depression: int = -6,
            accumulator_potentiation: int = 6,
            mean_pre_window: float = 35.0, mean_post_window: float = 35.0,
            dual_fsm: bool = True,
            A_plus: float = 0.01, A_minus: float = 0.01):
        """
        :param accumulator_depression:
        :param accumulator_potentiation:
        :param mean_pre_window:
        :param mean_post_window:
        :param dual_fsm:
        :param A_plus: :math:`A^+`
        :param A_minus: :math:`A^-`
        """
        super().__init__(SynapseStructureWeightAccumulator())
        self.__accumulator_depression_plus_one = accumulator_depression + 1
        self.__accumulator_potentiation_minus_one = \
            accumulator_potentiation - 1
        self.__mean_pre_window = mean_pre_window
        self.__mean_post_window = mean_post_window
        self.__dual_fsm = dual_fsm
        self.__a_plus = A_plus
        self.__a_minus = A_minus

    @property
    def A_plus(self) -> float:
        r"""
        :math:`A^+`
        """
        return self.__a_plus

    @A_plus.setter
    def A_plus(self, new_value: float) -> None:
        self.__a_plus = new_value

    @property
    def A_minus(self) -> float:
        r"""
        :math:`A^-`
        """
        return self.__a_minus

    @A_minus.setter
    def A_minus(self, new_value: float) -> None:
        self.__a_minus = new_value

    @overrides(AbstractTimingDependence.is_same_as)
    def is_same_as(self, timing_dependence: AbstractTimingDependence) -> bool:
        if not isinstance(timing_dependence, TimingDependenceRecurrent):
            return False
        # pylint: disable=protected-access
        return self._character() == timing_dependence._character()

    def _character(self) -> object:
        """
        Two instances of this class are the same if their characterisation is
        the same.
        """
        return (self.__accumulator_depression_plus_one,
                self.__accumulator_potentiation_minus_one,
                self.__mean_pre_window, self.__mean_post_window)

    @property
    def vertex_executable_suffix(self) -> str:
        """
        The suffix to be appended to the vertex executable for this rule.
        """
        if self.__dual_fsm:
            return "recurrent_dual_fsm"
        return "recurrent_pre_stochastic"

    @property
    def pre_trace_n_bytes(self) -> int:
        """
        The number of bytes used by the pre-trace of the rule per neuron.
        """
        # When using the separate FSMs, pre-trace contains window length,
        # otherwise it's in the synapse
        return BYTES_PER_SHORT if self.__dual_fsm else 0

    @overrides(AbstractTimingDependence.get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(self) -> int:
        # 2 * 32-bit parameters
        # 2 * LUTS with STDP_FIXED_POINT_ONE * 16-bit entries
        return (2 * BYTES_PER_WORD) + (
            2 * STDP_FIXED_POINT_ONE * BYTES_PER_SHORT)

    @property
    def n_weight_terms(self) -> int:
        """
        The number of weight terms expected by this timing rule.
        """
        return 1

    @overrides(AbstractTimingDependence.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, global_weight_scale: float,
            synapse_weight_scales: NDArray[floating]) -> None:
        # Write parameters
        spec.write_value(data=self.__accumulator_depression_plus_one,
                         data_type=DataType.INT32)
        spec.write_value(data=self.__accumulator_potentiation_minus_one,
                         data_type=DataType.INT32)

        # Convert mean times into machine timesteps
        time_step_per_ms = SpynnakerDataView.get_simulation_time_step_per_ms()

        mean_pre_timesteps = float(self.__mean_pre_window * time_step_per_ms)
        mean_post_timesteps = float(self.__mean_post_window * time_step_per_ms)

        # Write lookup tables
        self._write_exp_dist_lut(spec, mean_pre_timesteps)
        self._write_exp_dist_lut(spec, mean_post_timesteps)

    @staticmethod
    def _write_exp_dist_lut(spec: DataSpecificationBase, mean: float) -> None:
        indices = numpy.arange(STDP_FIXED_POINT_ONE)
        inv_cdf = numpy.log(1.0 - indices/float(STDP_FIXED_POINT_ONE)) * -mean
        spec.write_array(
            inv_cdf.astype(numpy.uint16), data_type=DataType.UINT16)

    @overrides(AbstractTimingDependence.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        return self.__PARAM_NAMES



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This contains functions and classes for handling external devices such as the
PushBot (https://spinnakermanchester.github.io/docs/push_bot/).

.. note::
    When using external devices, it is normally important to configure your
    SpiNNaker system to run in real-time mode, which usually reduces numerical
    accuracy to gain performance.
"""
import os
from typing import Optional, Tuple
from spinn_utilities.socket_address import SocketAddress
from spinnman.messages.eieio import EIEIOType
from spinn_front_end_common.abstract_models import (
    AbstractSendMeMulticastCommandsVertex)
from spinn_front_end_common.utilities.utility_objs import (
    LivePacketGatherParameters)
from spynnaker.pyNN.external_devices_models import (
    AbstractEthernetController, AbstractEthernetSensor,
    ArbitraryFPGADevice, ExternalFPGARetinaDevice,
    MunichMotorDevice, MunichRetinaDevice, ExternalDeviceLifControl,
    SPIFRetinaDevice, ICUBRetinaDevice, SPIFOutputDevice, SPIFInputDevice)
from spynnaker.pyNN import model_binaries
from spynnaker.pyNN.connections import (
    EthernetCommandConnection, EthernetControlConnection,
    SpynnakerLiveSpikesConnection, SpynnakerPoissonControlConnection,
    SPIFLiveSpikesConnection)
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.external_devices_models.push_bot.control import (
    PushBotLifEthernet, PushBotLifSpinnakerLink)
from spynnaker.pyNN.external_devices_models.push_bot.spinnaker_link import (
    PushBotSpiNNakerLinkRetinaDevice,
    PushBotSpiNNakerLinkLaserDevice, PushBotSpiNNakerLinkLEDDevice,
    PushBotSpiNNakerLinkMotorDevice, PushBotSpiNNakerLinkSpeakerDevice)
from spynnaker.pyNN.external_devices_models.push_bot.ethernet import (
    PushBotEthernetLaserDevice, PushBotEthernetLEDDevice,
    PushBotEthernetMotorDevice, PushBotEthernetRetinaDevice,
    PushBotEthernetSpeakerDevice)
from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotLaser, PushBotLED, PushBotMotor, PushBotRetinaResolution,
    PushBotSpeaker, PushBotRetinaViewer)
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol
from spynnaker.pyNN.spynnaker_external_device_plugin_manager import (
    SpynnakerExternalDevicePluginManager as
    Plugins)
from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel
from spynnaker.pyNN.models.populations.population import (
    Population, _CellTypeArg)
from spynnaker.pyNN.models.utility_models.spike_injector import (
    SpikeInjector as ExternalDeviceSpikeInjector)
from spynnaker.pyNN import protocols
from spynnaker.pyNN.spinnaker import SpiNNaker
from spynnaker.pyNN.models.neuron import PopulationVertex


# useful functions
add_database_socket_address = Plugins.add_database_socket_address
activate_live_output_to = Plugins.activate_live_output_to
activate_live_output_for = Plugins.activate_live_output_for
add_poisson_live_rate_control = Plugins.add_poisson_live_rate_control

SpynnakerDataView.register_binary_search_path(
    os.path.dirname(model_binaries.__file__))
spynnaker_external_devices = Plugins()

__all__ = [
    "EIEIOType",

    # General Devices
    "ExternalFPGARetinaDevice",
    "MunichRetinaDevice", "MunichMotorDevice", "ArbitraryFPGADevice",
    "PushBotRetinaViewer", "ExternalDeviceLifControl", "SPIFRetinaDevice",
    "ICUBRetinaDevice", "SPIFOutputDevice", "SPIFInputDevice",

    # PushBot Parameters
    "MunichIoSpiNNakerLinkProtocol",
    "PushBotLaser", "PushBotLED", "PushBotMotor", "PushBotSpeaker",
    "PushBotRetinaResolution",

    # PushBot Ethernet Parts
    "PushBotLifEthernet", "PushBotEthernetLaserDevice",
    "PushBotEthernetLEDDevice", "PushBotEthernetMotorDevice",
    "PushBotEthernetSpeakerDevice", "PushBotEthernetRetinaDevice",

    # PushBot SpiNNaker Link Parts
    "PushBotLifSpinnakerLink", "PushBotSpiNNakerLinkLaserDevice",
    "PushBotSpiNNakerLinkLEDDevice", "PushBotSpiNNakerLinkMotorDevice",
    "PushBotSpiNNakerLinkSpeakerDevice", "PushBotSpiNNakerLinkRetinaDevice",

    # Connections
    "SpynnakerLiveSpikesConnection",
    "SpynnakerPoissonControlConnection",
    "SPIFLiveSpikesConnection",

    # Provided functions
    "activate_live_output_for",
    "activate_live_output_to",
    "SpikeInjector",
    "register_database_notification_request",
    "run_forever",
    "add_poisson_live_rate_control",
    "protocols"
]
# Cache of the simulator provided by pyNN/__init__py
__simulator: Optional[SpiNNaker] = None


def run_forever(sync_time: float = 0.0) -> None:
    """
    Supports running forever in PyNN 0.8/0.9 format.

    Returns only when the application has started running on the SpiNNaker
        platform

    :param sync_time:
        The time in milliseconds after which to pause before the host must
        continue the simulation.

    """
    SpynnakerDataView.check_user_can_act()
    assert __simulator is not None, "no simulator set up"
    __simulator.run(None, sync_time)


def run_sync(run_time: float, sync_time: float) -> None:
    """
    Run in steps of the given number of milliseconds pausing between
    for a signal to be sent from the host.

    :param run_time: The time in milliseconds to run the simulation for
    :param sync_time: The time in milliseconds to pause before allowing
    """
    SpynnakerDataView.check_user_can_act()
    assert __simulator is not None, "no simulator set up"
    __simulator.run(run_time, sync_time)


def continue_simulation() -> None:
    """
    Continue a synchronised simulation.
    """
    SpynnakerDataView.check_valid_simulator()
    assert __simulator is not None, "no simulator set up"
    __simulator.continue_simulation()


def request_stop() -> None:
    """
    Request a stop in the simulation without a complete stop.  Will stop
    after the next auto-pause-and-resume cycle.
    """
    SpynnakerDataView.check_valid_simulator()
    assert __simulator is not None, "no simulator set up"
    __simulator.stop_run()


def register_database_notification_request(
        hostname: str, notify_port: int, ack_port: int) -> None:
    """
    Adds a socket system which is registered with the notification protocol.

    :param hostname: hostname to connect to
    :param notify_port: UDP port number for the notify command
    :param ack_port: UDP port number for the acknowledge command
    """
    SpynnakerDataView.add_database_socket_address(
        SocketAddress(hostname, notify_port, ack_port))


# Store the connection to be used by multiple users
__ethernet_control_connection: Optional[EthernetControlConnection] = None


def __vtx(population: Population) -> Tuple[
        PopulationVertex, AbstractEthernetController, str]:
    vertex = population._vertex  # pylint: disable=protected-access
    if isinstance(vertex, PopulationVertex):
        v = vertex
    else:
        raise TypeError(
            "Vertex must be an instance of PopulationVertex")
    if isinstance(vertex, AbstractEthernetController):
        c = vertex
    else:
        raise TypeError(
            "Vertex must be an instance of AbstractEthernetController")
    if vertex.label is None:
        raise ValueError("Vertex must be labelled")
    return v, c, vertex.label


def EthernetControlPopulation(
        n_neurons: int, model: _CellTypeArg, label: Optional[str] = None,
        local_host: Optional[str] = None, local_port: Optional[int] = None,
        database_notify_port_num: Optional[int] = None,
        database_ack_port_num: Optional[int] = None) -> Population:
    # pylint: disable=invalid-name
    """
    Create a PyNN population that can be included in a network to
    control an external device which is connected to the host.

    :param n_neurons: The number of neurons in the control population
    :param model:
        Class of a model that creates a vertex of type
        :py:class:`AbstractEthernetController`
    :param label: An optional label for the population
    :param local_host:
        The optional local host IP address to listen on for commands
    :param local_port: The optional local port to listen on for commands
    :param database_ack_port_num:
        The optional port to which responses to the database notification
        protocol are to be sent
    :param database_notify_port_num:
        The optional port to which notifications from the database
        notification protocol are to be sent
    :return:
        A pyNN Population which can be used as the target of a Projection.

        .. note::
            The Population can also be used as the source of a
            Projection, but it might not send spikes.
    :raises TypeError: If an invalid model class is used.
    """
    # pylint: disable=global-statement
    population = Population(n_neurons, model, label=label)
    vertex, aec, vertex_label = __vtx(population)
    translator = aec.get_message_translator()
    live_packet_gather_label = "EthernetControlReceiver"
    global __ethernet_control_connection
    if __ethernet_control_connection is None:
        __ethernet_control_connection = EthernetControlConnection(
            translator, vertex_label, live_packet_gather_label, local_host,
            local_port)
        Plugins.add_database_socket_address(
            __ethernet_control_connection.local_ip_address,
            __ethernet_control_connection.local_port, database_ack_port_num)
    else:
        __ethernet_control_connection.add_translator(vertex_label, translator)
    devices_with_commands = [
        device for device in aec.get_external_devices()
        if isinstance(device, AbstractSendMeMulticastCommandsVertex)]
    if devices_with_commands:
        ethernet_command_connection = EthernetCommandConnection(
            translator, devices_with_commands, local_host,
            database_notify_port_num)
        Plugins.add_database_socket_address(
            ethernet_command_connection.local_ip_address,
            ethernet_command_connection.local_port, database_ack_port_num)
    params = LivePacketGatherParameters(
        port=__ethernet_control_connection.local_port,
        hostname=__ethernet_control_connection.local_ip_address,
        message_type=EIEIOType.KEY_PAYLOAD_32_BIT,
        payload_as_time_stamps=False, use_payload_prefix=False,
        label=live_packet_gather_label)
    Plugins.update_live_packet_gather_tracker(
        vertex, params, aec.get_outgoing_partition_ids())
    return population


def EthernetSensorPopulation(
        device: AbstractEthernetSensor, local_host: Optional[str] = None,
        database_notify_port_num: Optional[int] = None,
        database_ack_port_num: Optional[int] = None) -> Population:
    # pylint: disable=invalid-name
    """
    Create a pyNN population which can be included in a network to
    receive spikes from a device connected to the host.

    :param device: The sensor model
    :param local_host:
        The optional local host IP address to listen on for database
        notification
    :param database_ack_port_num:
        The optional port to which responses to the database notification
        protocol are to be sent
    :param database_notify_port_num:
        The optional port to which notifications from the database
        notification protocol are to be sent
    :return:
        A pyNN Population which can be used as the source of a Projection.

        .. note::
            The Population cannot be used as the target of a Projection.
    :raises TypeError: If an invalid model class is used.
    """
    if not isinstance(device, AbstractEthernetSensor):
        raise TypeError(
            "Device must be an instance of AbstractEthernetSensor")
    injector_params = dict(device.get_injector_parameters())

    population = Population(
        device.get_n_neurons(), SpikeInjector(notify=False),
        label=device.get_injector_label(),
        additional_parameters=injector_params)
    if isinstance(device, AbstractSendMeMulticastCommandsVertex):
        cmd_conn = EthernetCommandConnection(
            device.get_translator(), [device], local_host,
            database_notify_port_num)
        Plugins.add_database_socket_address(
            cmd_conn.local_ip_address, cmd_conn.local_port,
            database_ack_port_num)
    db_conn = device.get_database_connection()
    if db_conn is not None:
        Plugins.add_database_socket_address(
            db_conn.local_ip_address, db_conn.local_port,
            database_ack_port_num)
    return population


def SpikeInjector(
        notify: bool = True, database_notify_host: Optional[str] = None,
        database_notify_port_num: Optional[int] = None,
        database_ack_port_num: Optional[int] = None) -> AbstractPyNNModel:
    # pylint: disable=invalid-name
    """
    Supports creating a spike injector that can be added to the
    application graph.

    :param notify: Whether to register for notifications
    :param database_notify_host: the hostname for the device which is
        listening to the database notification.
    :param database_ack_port_num: the port number to which a external device
        will acknowledge that they have finished reading the database and are
        ready for it to start execution
    :param database_notify_port_num: The port number to which a external
        device will receive the database is ready command
    :return: The spike injector model object that can be placed in a pyNN
        :py:class:`~spynnaker.pyNN.models.populations.Population`.
    """
    if notify:
        Plugins.add_database_socket_address(
            database_notify_host, database_notify_port_num,
            database_ack_port_num)
    return ExternalDeviceSpikeInjector()


def _set_simulator(simulator: SpiNNaker) -> None:
    """
    Should only be called by pyNN/__init__py setup method.

    Any other uses is not supported.

    :param simulator:
    """
    global __simulator  # pylint: disable=global-statement
    __simulator = simulator



# Copyright (c) 2021 The University of Manchester
# Based on work Copyright (c) The University of Sussex,
# Garibaldi Pineda Garcia, James Turner, James Knight and Thomas Nowotny
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from collections.abc import (Iterable, Sequence)
from typing import (
    List, Optional, Sequence as TSequence, Tuple, Union,
    cast, overload, TYPE_CHECKING)

import numpy
from numpy import floating, float64, integer, int16, uint16, uint32
from numpy.typing import NDArray

from pyNN.random import RandomDistribution

from spinn_utilities.overrides import overrides

from pacman.model.graphs.abstract_vertex import AbstractVertex
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.common import Slice
from pacman.model.graphs.machine import MachineVertex

from spinn_front_end_common.utilities.constants import (
    BYTES_PER_SHORT, BYTES_PER_WORD)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.common.local_only_2d_common import get_div_const

from .abstract_connector import AbstractConnector

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)

#: The size of the connector struct in bytes
CONNECTOR_CONFIG_SIZE = (10 * BYTES_PER_SHORT) + (4 * BYTES_PER_WORD)


_Weights = Union[
    int, float, List[Union[int, float]], Tuple[Union[int, float], ...],
    NDArray[float64], RandomDistribution]
_Shape = Union[int, Tuple[int, int], None]
_Padding = Union[bool, _Shape]


class ConvolutionConnector(AbstractConnector):
    """
    A 2D connector that centres on a post neuron.

    Special connector which dynamically maps a 2D kernel over a 2D plane,
    so that when the kernel is centred on a post-neuron,
    that post-neuron receives input from each pre-neuron
    that the kernel touches,
    and the weights of those pre-post connections
    are then the values of the kernel.
    """

    __slots__ = (
        "__kernel_weights",
        "__strides",
        "__padding_shape",
        "__pool_shape",
        "__pool_stride",
        "__positive_receptor_type",
        "__negative_receptor_type",
        "__filter_edges"
    )

    def __init__(self, kernel_weights: _Weights,
                 kernel_shape: _Shape = None,
                 strides: _Shape = None, padding: _Padding = None,
                 pool_shape: _Shape = None, pool_stride: _Shape = None,
                 positive_receptor_type: str = "excitatory",
                 negative_receptor_type: str = "inhibitory",
                 safe: bool = True, verbose: bool = False,
                 callback: None = None, filter_edges: bool = True):
        """
        :param kernel_weights:
            The synaptic strengths, shared by neurons in the post population.
            Can be:

            * single value: `kernel_shape` must be provided;
              the same value will be used for all weights
            * simple list: `kernel_shape` must be provided; the list must
              be sized shape width * height
            * 2D list: If `kernel_shape` is provided, it must match
            * :py:class:`~numpy.ndarray`: As above for simple or 2D list
            * :py:class:`~spynnaker.pyNN.RandomDistribution`:
              `kernel_shape` must be provided; weights will be drawn from the
              distribution
        :param kernel_shape:
            The shape of the kernel if it cannot be determined from
            `kernel_weights`. If a single value is provided, a square kernel
            will be assumed.  If two values are provided, it will be assumed to
            be (n_rows, n_columns)
        :param strides:
            Spatial sampling frequency, jumps between the post neurons.
            This matches the meaning of standard ML packages.  If a single
            value is provided, the same stride will be used for rows and
            columns.  If two values are provided it will be assumed to be
            (stride_rows, stride_columns)
        :param padding:
            How many 'extra pixels' around the pre-population will be added,
            only zero-valued pixels are currently supported.  If a single
            value is provided, the same padding will be used for rows and
            columns.  If two values are provided it will be assumed to be
            `(padding_rows, padding_columns)`.  If True, automatic padding will
            be used based on the kernel shape.  If False or `None`, no padding
            will be used.
        :param pool_shape:
            Area of pooling, only average pooling is supported (and seems to
            make sense). If a single value is provided, the pooling area will
            be square.  If two values are provided it will be assumed to be
            `(pooling_rows, pooling_columns)`.
        :param pool_stride:
            Jumps between pooling regions. If a single value is provided, the
            same stride will be used for rows and columns.  If two values are
            provided it will be assumed to be `(stride_rows, stride_columns)`
        :param positive_receptor_type:
            The receptor type to add the positive weights to.  By default this
            is "``excitatory``".
        :param negative_receptor_type:
            The receptor type to add the negative weights to.  By default this
            is "``inhibitory``".
        :param safe:
        :param verbose:
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        :param filter_edges:
            Whether to filter the edges based on connectivity or not; filtered
            means that the receiving cores will receive fewer packets, whereas
            non-filtered means that receiving cores will receive all packets
            whether relevant or not.  However non-filtered may be more
            efficient in the routing tables, so may be needed if routing
            compression doesn't work.
        """
        super().__init__(safe=safe, callback=callback, verbose=verbose)

        self.__kernel_weights = self.__decode_kernel(
            kernel_weights, kernel_shape)
        self.__padding_shape = self.__decode_padding(padding)
        self.__filter_edges = filter_edges

        self.__strides: NDArray[integer]
        if strides is None:
            self.__strides = numpy.array((1, 1), dtype=numpy.uint32)
        else:
            self.__strides = self.__to_2d_shape(strides, "strides")
        self.__pool_shape = self.__to_2d_shape(pool_shape, "pool_shape")
        self.__pool_stride = self.__to_2d_shape(pool_stride, "pool_stride")
        if self.__pool_stride is None:
            self.__pool_stride = self.__pool_shape
        if self.__pool_shape is not None:
            self.__kernel_weights /= numpy.prod(self.__pool_shape)

        self.__positive_receptor_type = positive_receptor_type
        self.__negative_receptor_type = negative_receptor_type

    @property
    def positive_receptor_type(self) -> str:
        """
        The receptor type to add the positive weights to.
        """
        return self.__positive_receptor_type

    @property
    def negative_receptor_type(self) -> str:
        """
         The receptor type to add the negative weights to.
        """
        return self.__negative_receptor_type

    @property
    def kernel_weights(self) -> NDArray[float64]:
        """
        The weights for this connection.
        """
        return self.__kernel_weights

    def __get_kernel_shape(self, shape: _Shape) -> Tuple[int, int]:
        if shape is None:
            raise SynapticConfigurationException(
                "kernel_shape must be provided")
        if numpy.isscalar(shape):
            assert isinstance(shape, int)
            return (shape, shape)
        if isinstance(shape, tuple) and len(shape) == 2:
            return shape
        raise SynapticConfigurationException(f"Unknown kernel_shape: {shape}")

    def __decode_kernel(self, w: _Weights, shape: _Shape) -> NDArray[float64]:
        if isinstance(w, (int, float)):
            _shape = self.__get_kernel_shape(shape)
            return numpy.full(_shape, w, dtype=float64)
        elif isinstance(w, (Sequence, numpy.ndarray)):
            if all(isinstance(lst, (Sequence, numpy.ndarray)) for lst in w):
                ws = cast(TSequence[TSequence[float]], w)
                len0 = len(ws[0])
                # 2D list
                if not all(len(lst) == len0 for lst in ws):
                    raise SynapticConfigurationException(
                        "kernel_weights must be a 2D array with every row the"
                        " same length")
                return numpy.array(w, dtype=float64)
            else:
                # 1D list
                _shape = self.__get_kernel_shape(shape)
                return numpy.array(w, dtype=float64).reshape(_shape)
        elif isinstance(w, RandomDistribution):
            _shape = self.__get_kernel_shape(shape)
            return numpy.array(
                w.next(numpy.prod(_shape)), dtype=float64).reshape(_shape)
        else:
            raise SynapticConfigurationException(
                f"Unknown combination of kernel_weights ({w}) and"
                f" kernel_shape ({shape})")

    @overload
    @staticmethod
    def __to_2d_shape(shape: Union[int, Tuple[int, int]],
                      param_name: str) -> NDArray[integer]:
        ...

    @overload
    @staticmethod
    def __to_2d_shape(shape: None, param_name: str) -> None:
        ...

    @staticmethod
    def __to_2d_shape(shape: _Shape, param_name: str) -> Optional[
            NDArray[integer]]:
        if shape is None:
            return None
        if numpy.isscalar(shape):
            return numpy.array([shape, shape], dtype=uint32)
        assert isinstance(shape, tuple)
        if len(shape) == 1:
            return numpy.array([shape[0], 1], dtype=uint32)
        elif len(shape) == 2:
            return numpy.array(shape, dtype=uint32)
        raise SynapticConfigurationException(
            f"{param_name} must be an int or a tuple(int, int)")

    def __decode_padding(self, padding: _Padding) -> NDArray[integer]:
        if isinstance(padding, (int, Iterable)):
            return self.__to_2d_shape(padding, "padding")
        elif padding is None or padding is False:
            return numpy.zeros(2, dtype=uint32)
        elif padding:
            return self.__kernel_weights.shape // 2
        else:
            raise SynapticConfigurationException(
                f"Unrecognized padding {padding}")

    def get_post_shape(self, shape: Tuple[int, ...]) -> Tuple[int, ...]:
        """
        :returns: The shape of the post image given the pre-image shape.
        """
        _shape = numpy.array(shape)
        if self.__pool_shape is not None:
            _shape = _shape // self.__pool_stride

        kernel_shape = numpy.array(self.__kernel_weights.shape)
        post_shape = (_shape - (kernel_shape - 1) +
                      (2 * self.__padding_shape))

        return tuple(int(i) for i in numpy.clip(
            post_shape // self.__strides, 1, numpy.inf).astype(
            numpy.uint32))

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        pre = application_edge.pre_vertex
        post = application_edge.post_vertex
        if len(pre.atoms_shape) != 2 or len(post.atoms_shape) != 2:
            raise ConfigurationException(
                "The ConvolutionConnector only works where the Populations"
                " of a Projection are both 2D.  Please ensure that both the"
                " Populations use a Grid2D structure.")
        expected_post_shape = tuple(self.get_post_shape(pre.atoms_shape))
        if expected_post_shape != post.atoms_shape:
            raise ConfigurationException(
                f"With a source population with shape {pre.atoms_shape}, "
                "for a Convolution connector with the given parameters, "
                "the post-population must have a shape "
                f"{expected_post_shape}")
        if post.get_synapse_id_by_target(
                self.__positive_receptor_type) is None:
            raise ConfigurationException(
                "The post population doesn't have a synaptic receptor type of"
                f" {self.__positive_receptor_type}")
        if post.get_synapse_id_by_target(
                self.__negative_receptor_type) is None:
            raise ConfigurationException(
                "The post population doesn't have a synaptic receptor type of"
                f" {self.__negative_receptor_type}")
        if not isinstance(synapse_info.delays, float):
            raise ConfigurationException(
                "The ConvolutionConnector only supports simple uniform delays")

    @staticmethod
    def __delay(synapse_info: SynapseInformation) -> float:
        # Checked by validate_connection above
        return cast(float, synapse_info.delays)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self.__delay(synapse_info)

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self.__delay(synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        if min_delay is not None and max_delay is not None:
            if not (min_delay <= self.__delay(synapse_info) <= max_delay):
                return 0
        w, h = self.__kernel_weights.shape
        return numpy.clip(w * h, 0, n_post_atoms)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        w, h = self.__kernel_weights.shape
        return numpy.clip(w * h, 0, synapse_info.n_pre_neurons)

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        return float(numpy.amax(self.__kernel_weights))

    @overrides(AbstractConnector.get_connected_vertices)
    def get_connected_vertices(
            self, s_info: SynapseInformation,
            source_vertex: ApplicationVertex,
            target_vertex: ApplicationVertex) -> Sequence[
                Tuple[MachineVertex, Sequence[AbstractVertex]]]:
        if not self.__filter_edges:
            return super(ConvolutionConnector, self).get_connected_vertices(
                s_info, source_vertex, target_vertex)
        pre_vertices = numpy.array(
            source_vertex.splitter.get_out_going_vertices(s_info.partition_id))
        post_slice_ranges = self.__pre_as_post_slice_ranges(
            m_vertex.vertex_slice for m_vertex in pre_vertices)
        hlf_k_w, hlf_k_h = numpy.array(self.__kernel_weights.shape) // 2

        connected: List[Tuple[MachineVertex, List[MachineVertex]]] = []
        for post in target_vertex.splitter.get_in_coming_vertices(
                s_info.partition_id):
            post_slice = post.vertex_slice
            post_slice_x = post_slice.get_slice(0)
            post_slice_y = post_slice.get_slice(1)

            # Get ranges allowed in post
            min_x = post_slice_x.start - hlf_k_w
            max_x = (post_slice_x.stop + hlf_k_w) - 1
            min_y = post_slice_y.start - hlf_k_h
            max_y = (post_slice_y.stop + hlf_k_h) - 1

            # Test that the start coordinates are in range i.e. less than max
            start_in_range = numpy.logical_not(
                numpy.any(post_slice_ranges[:, 0] > [max_x, max_y], axis=1))
            # Test that the end coordinates are in range i.e. more than min
            end_in_range = numpy.logical_not(
                numpy.any(post_slice_ranges[:, 1] < [min_x, min_y], axis=1))
            # When both things are true, we have a vertex in range
            pre_in_range = pre_vertices[
                numpy.logical_and(start_in_range, end_in_range)]
            connected.append((post, list(pre_in_range)))
        return connected

    def __pre_as_post_slice_ranges(
            self, slices: Iterable[Slice]) -> NDArray[integer]:
        """
        Convert a generator of (multi-dimensional) pre-slices into an array of
        post-slices.
        """
        pre_slices = ((s.get_slice(0), s.get_slice(1)) for s in slices)
        coords = numpy.array([
            ((px.start, py.start), (px.stop - 1, py.stop - 1))
            for px, py in pre_slices])
        if self.__pool_stride is not None:
            coords //= self.__pool_stride

        kernel_shape = numpy.array(self.__kernel_weights.shape)
        coords = coords - kernel_shape // 2 + self.__padding_shape
        coords //= self.__strides
        return coords

    @property
    def kernel_n_bytes(self) -> int:
        """
        Size of the weights in bytes
        """
        n_weights = self.__kernel_weights.size
        return n_weights * BYTES_PER_SHORT

    @property
    def kernel_n_weights(self) -> int:
        """
        Size of the weights.
        """
        return self.__kernel_weights.size

    @property
    def parameters_n_bytes(self) -> int:
        """
        The size of the connector struct in bytes
        """
        return CONNECTOR_CONFIG_SIZE

    def get_local_only_data(
            self, app_edge: ProjectionApplicationEdge, local_delay: int,
            delay_stage: int, weight_index: int) -> NDArray[uint32]:
        """
        :param app_edge:
        :param local_delay:
        :param delay_stage:
        :param weight_index:
        :returns: The local only data
        """
        # Get info about things
        kernel_shape = self.__kernel_weights.shape
        ps_x, ps_y = 1, 1
        if self.__pool_stride is not None:
            ps_x, ps_y = self.__pool_stride

        # Do a new list for remaining connector details as uint16s
        pos_synapse_type = app_edge.post_vertex.get_synapse_id_by_target(
            self.__positive_receptor_type)
        neg_synapse_type = app_edge.post_vertex.get_synapse_id_by_target(
            self.__negative_receptor_type)

        # Produce the values needed
        short_values = numpy.array([
            kernel_shape[1], kernel_shape[0],
            self.__padding_shape[1], self.__padding_shape[0],
            pos_synapse_type, neg_synapse_type, delay_stage, local_delay,
            weight_index, 0], dtype=uint16)
        long_values = numpy.array([
            get_div_const(self.__strides[1]), get_div_const(self.__strides[0]),
            get_div_const(ps_y), get_div_const(ps_x)], dtype=uint32)
        return numpy.concatenate((short_values.view(uint32), long_values))

    def get_encoded_kernel_weights(
            self, app_edge: ProjectionApplicationEdge,
            weight_scales: NDArray[floating]) -> NDArray[int16]:
        """
        :param app_edge:
        :param weight_scales:
        :returns: Encoded weights with weight scaling.
        """
        encoded_kernel_weights = self.__kernel_weights.flatten()
        neg_weights = encoded_kernel_weights < 0
        pos_weights = encoded_kernel_weights > 0
        pos_synapse_type = app_edge.post_vertex.get_synapse_id_by_target(
            self.__positive_receptor_type)
        neg_synapse_type = app_edge.post_vertex.get_synapse_id_by_target(
            self.__negative_receptor_type)
        encoded_kernel_weights[neg_weights] *= weight_scales[neg_synapse_type]
        encoded_kernel_weights[pos_weights] *= weight_scales[pos_synapse_type]
        return numpy.round(encoded_kernel_weights).astype(int16)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from abc import abstractmethod
from enum import Enum
from typing import TYPE_CHECKING

import numpy
from numpy import uint32
from numpy.typing import NDArray

from pyNN.random import RandomDistribution

from spinn_utilities.abstract_base import AbstractBase
from spinn_utilities.overrides import overrides

from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractConnector)
from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.common.param_generator_data import (
    param_generator_params, param_generator_params_size_in_bytes,
    param_generator_id, is_param_generatable)
from spynnaker.pyNN.types import (DELAYS, WEIGHTS)
from spynnaker.pyNN.utilities.utility_calls import check_rng

from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)


class ConnectorIDs(Enum):
    """
    Hashes of the connection generators supported by the synapse expander
    """
    ONE_TO_ONE_CONNECTOR = 0
    ALL_TO_ALL_CONNECTOR = 1
    FIXED_PROBABILITY_CONNECTOR = 2
    FIXED_TOTAL_NUMBER_CONNECTOR = 3
    FIXED_NUMBER_PRE_CONNECTOR = 4
    FIXED_NUMBER_POST_CONNECTOR = 5
    KERNEL_CONNECTOR = 6
    ALL_BUT_ME_CONNECTOR = 7
    ONE_TO_ONE_OFFSET_CONNECTOR = 8


class AbstractGenerateConnectorOnMachine(
        AbstractConnector, metaclass=AbstractBase):
    """
    Indicates that the connectivity can be generated on the machine.
    """

    __slots__ = ()

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        # If we can't generate on machine, we must be able to generate on host
        if not self.generate_on_machine(synapse_info):
            if not isinstance(self, AbstractGenerateConnectorOnHost):
                raise SynapticConfigurationException(
                    "The parameters of this connection do not allow it to be"
                    " generated on the machine, but the connector cannot"
                    " be generated on host!")

    def generate_on_machine(self, synapse_info: SynapseInformation) -> bool:
        """
        Determine if this instance can generate on the machine.

        Default implementation returns True if the weights and delays can
        be generated on the machine

        :param synapse_info: The synapse information
        :returns: True if this instance can generate on the machine.
        """
        if (not is_param_generatable(synapse_info.weights) or
                not is_param_generatable(synapse_info.delays)):
            return False
        if isinstance(synapse_info.weights, RandomDistribution):
            check_rng(synapse_info.weights.rng, "RandomDistribution in weight")
        if isinstance(synapse_info.delays, RandomDistribution):
            check_rng(synapse_info.delays.rng, "RandomDistribution in delay")
        return True

    def gen_weights_id(self, weights: WEIGHTS) -> int:
        """
        :param weights:
        :returns: The id of the weight generator on the machine.
        """
        return param_generator_id(weights)

    def gen_weights_params(self, weights: WEIGHTS) -> NDArray[uint32]:
        """
        :param weights:
        :returns: The parameters of the weight generator on the machine.
        """
        return param_generator_params(weights)

    def gen_weight_params_size_in_bytes(self, weights:  WEIGHTS) -> int:
        """
        :param weights:
        :returns: The size of the weight parameters in bytes.
        """
        return param_generator_params_size_in_bytes(weights)

    def gen_delays_id(self, delays: DELAYS) -> int:
        """
        :param delays:
        :returns: The id of the delay generator on the machine.
        """
        return param_generator_id(delays)

    def gen_delay_params(self, delays: DELAYS) -> NDArray[uint32]:
        """
        :param delays:
        :returns: The parameters of the delay generator on the machine.
        """
        return param_generator_params(delays)

    def gen_delay_params_size_in_bytes(self, delays: DELAYS) -> int:
        """
        :param delays:
        :returns: The size of the delay parameters in bytes.
        """
        return param_generator_params_size_in_bytes(delays)

    @property
    @abstractmethod
    def gen_connector_id(self) -> int:
        """
        The ID of the connection generator on the machine.
        """
        raise NotImplementedError

    def gen_connector_params(
            self, synapse_info: SynapseInformation) -> NDArray[uint32]:
        """
        :param synapse_info: The synaptic information
        :returns: The parameters of the on machine generation.
        """
        _ = synapse_info
        return numpy.zeros(0, dtype="uint32")

    @property
    def gen_connector_params_size_in_bytes(self) -> int:
        """
        The size of the connector parameters, in bytes.
        """
        return 0



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
from typing import Dict, Set, Tuple, Type, TypeVar, TYPE_CHECKING

from spinn_utilities.config_holder import (
    get_config_str_or_none, get_report_path)
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from pacman.model.graphs.application import ApplicationVertex
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.models.neural_projections import ProjectionApplicationEdge
if TYPE_CHECKING:
    import graphviz as gv  # type: ignore[import]

logger = FormatAdapter(logging.getLogger(__name__))
_RE = TypeVar("_RE", bound=RuntimeError)

CUTOFF = 100
_GRAPH_TITLE = "The graph of the network in graphical form"
_GRAPH_FORMAT = "png"


def _get_diagram(label: str) -> Tuple[gv.Digraph, Type[gv.ExecutableNotFound]]:
    # pylint: disable=import-error,import-outside-toplevel
    try:
        import graphviz
    except ImportError as e:
        raise SpynnakerException(
            "graphviz is required to use this report. "
            "Please install graphviz if you want to use this report."
            ) from e
    return (graphviz.Digraph(comment=label),
            graphviz.ExecutableNotFound)


def spynnaker_neuron_graph_network_specification_report() -> None:
    """
    Produces a report describing the graph created from the neural
    populations and projections.
    """
    # create holders for data
    dot_diagram, exe_not_found_exn = _get_diagram(_GRAPH_TITLE)

    graph_format = get_config_str_or_none("Reports", "network_graph_format")
    if graph_format is None:
        graph_format = _GRAPH_FORMAT
        if (SpynnakerDataView.get_n_vertices() +
                SpynnakerDataView.get_n_partitions()) > CUTOFF:
            logger.warning(
                "cfg write_network_graph ignored as network_graph_format "
                "is None and the network is big")
            return
    # build progress bar for the vertices, edges, and rendering
    progress = ProgressBar(
        SpynnakerDataView.get_n_vertices() +
        SpynnakerDataView.get_n_partitions(),
        "generating the graphical representation of the neural network")

    # write vertices into dot diagram
    vertex_ids = _generate_vertices(dot_diagram, progress)
    # write edges into dot diagram
    _generate_edges(dot_diagram, vertex_ids, progress)

    # write dot file and generate PDF
    file_to_output = get_report_path("path_network_graph")

    logger.info(f"rendering dot diagram {file_to_output}")
    try:
        dot_diagram.render(file_to_output, view=False, format=graph_format)
    except exe_not_found_exn:
        logger.exception("could not render diagram in {}", file_to_output)
    finally:
        progress.end()


def _generate_vertices(
        dot_diagram: gv.Digraph,
        progress: ProgressBar) -> Dict[ApplicationVertex, str]:
    """
    :return: the mapping from vertex to ID for the generated vertices
    """
    vertex_ids: Dict[ApplicationVertex, str] = dict()
    for vertex_counter, vertex in progress.over(
            enumerate(SpynnakerDataView.iterate_vertices()), False):
        # Arbitrary labels used inside dot
        vertex_id = str(vertex_counter)
        dot_diagram.node(
            vertex_id, f"{vertex.label} ({vertex.n_atoms} neurons)")
        vertex_ids[vertex] = vertex_id
    return vertex_ids


def _generate_edges(
        dot_diagram: gv.Digraph, vertex_ids: Dict[ApplicationVertex, str],
        progress: ProgressBar) -> None:
    for partition in progress.over(
            SpynnakerDataView.iterate_partitions(), False):
        for edge in partition.edges:
            source_vertex_id = vertex_ids[edge.pre_vertex]
            dest_vertex_id = vertex_ids[edge.post_vertex]
            if isinstance(edge, ProjectionApplicationEdge):
                links: Set[str] = set()
                for synapse_info in edge.synapse_information:
                    links.add(str(synapse_info.connector))
                for synapse_str in links:
                    dot_diagram.edge(
                        source_vertex_id, dest_vertex_id, synapse_str)
            else:
                # Unlabelled edge
                dot_diagram.edge(source_vertex_id, dest_vertex_id)



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Union, cast

import numpy
from numpy import uint32, int32
from numpy.typing import NDArray
from typing_extensions import TypeAlias

from pyNN.random import RandomDistribution, available_distributions

from spinn_front_end_common.interface.ds import DataType
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD


#: The generator param type for each data type
_GENERATOR_TYPES = {
    DataType.S1615: 0,
    DataType.UINT32: 1,
    DataType.INT32: 2,
    DataType.U032: 3
}


def get_generator_type(data_type: DataType) -> int:
    """
    :param data_type:
    :return: The generator parameter type code for the given data type.
    :raises TypeError: If an unsupported data type is given
    """
    if data_type in _GENERATOR_TYPES:
        return _GENERATOR_TYPES[data_type]
    raise TypeError(f"Ungeneratable type {data_type}")


def type_has_generator(data_type: DataType) -> bool:
    """
    :param data_type:
    :return:
        Whether there is a generator parameter type code for the given data
        type.
    """
    return data_type in _GENERATOR_TYPES


#: ID of the constant parameter generator.
PARAM_TYPE_CONSTANT_ID = 0

#: IDs of the random parameter generators supported by the synapse expander.
PARAM_TYPE_BY_NAME = {
    "uniform": 1,
    "uniform_int": 1,
    "normal": 2,
    "normal_clipped": 3,
    "normal_clipped_to_boundary": 4,
    "exponential": 5,
    "exponential_clipped": 6
}

_ParamType: TypeAlias = Union[int, float, RandomDistribution]


def param_generator_id(value: _ParamType) -> int:
    """
    :param value: The value to examine the type of.
    :return: The ID of the on-chip generator that handles the value.
    :raises TypeError: If an value of an unsupported data type is given
    """
    # Scalars are fine on the machine
    if numpy.isscalar(value):
        return PARAM_TYPE_CONSTANT_ID

    # Only certain types of random distributions are supported for
    # generation on the machine
    if isinstance(value, RandomDistribution):
        if value.name in PARAM_TYPE_BY_NAME:
            return PARAM_TYPE_BY_NAME[value.name]

    raise TypeError(f"Ungeneratable parameter {value}")


def is_param_generatable(value: Any) -> bool:
    """
    :param value: The value to examine the type of.
    :return: Whether the value is of a type that can be generated on chip.
    """
    if isinstance(value, str):
        return False
    if numpy.isscalar(value):
        return True
    return (isinstance(value, RandomDistribution) and
            value.name in PARAM_TYPE_BY_NAME)


def param_generator_params(values: _ParamType) -> NDArray[uint32]:
    """
    :param values:
    :returns: The parameter generator parameters as a numpy array.
    """
    if numpy.isscalar(values):
        return numpy.array(
            [DataType.S1615.encode_as_int(cast(float, values))],
            dtype=int32).view(uint32)

    if isinstance(values, RandomDistribution):
        parameters = (
            values.parameters.get(param_name, None)
            for param_name in available_distributions[values.name])
        parameters = (
            DataType.S1615.max if param == numpy.inf
            else DataType.S1615.min if param == -numpy.inf else param
            for param in parameters if param is not None)
        params = [
            DataType.S1615.encode_as_int(param) for param in parameters]
        return numpy.array(params, dtype=int32).view(uint32)

    raise ValueError(f"Unexpected value {values}")


#: At most, there are 4 words as param generator parameters
MAX_PARAMS_BYTES = 4 * BYTES_PER_WORD


def param_generator_params_size_in_bytes(values: _ParamType) -> int:
    """
    :param values:
    :returns: The size of the parameter generator parameters in bytes.
    :raises TypeError: If `values` is of an unsupported data type
    """
    if numpy.isscalar(values):
        return BYTES_PER_WORD

    if isinstance(values, RandomDistribution):
        parameters = available_distributions[values.name]
        return len(parameters) * BYTES_PER_WORD

    raise ValueError(f"Unexpected value {values}")
